import os
import sys
import argparse
import subprocess

def parseArg():
    version = '3.0'
    parser = argparse.ArgumentParser(description='miRge3.0 (Comprehensive analysis of small RNA sequencing Data)',usage='miRge3.0 [options]',formatter_class=argparse.RawTextHelpFormatter,)
    if len(sys.argv)==1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    parser.add_argument('--version', action='version', version='%s'%(version))
    group = parser.add_argument_group("Options",description='''-s,    --samples            list of one or more samples separated by comma or a file with list of samples separated by new line (accepts *.fastq, *.fastq.gz) 
-db,   --mir-DB             the reference database of miRNA. Options: miRBase and miRGeneDB (Default: miRBase) 
-lib,  --libraries-path     the path to miRge libraries 
-on,   --organism-name      the organism name can be human, mouse, fruitfly, nematode, rat or zebrafish
-ex,   --crThreshold        the threshold of the proportion of canonical reads for the miRNAs to retain. Range for ex (0 - 0.5), (Default: 0.1)
-phr,  --phred64            phred64 format (Default: 33)
-spk,  --spikeIn            switch to annotate spike-ins if spike-in bowtie index files are located at the path of bowtie's index files (Default: off)
-ie,   --isoform-entropy    switch to calculate isomir entropy (default: off)
-cpu,  --threads            the number of processors to use for trimming, qc, and alignment (Default: 1)
-ai,   --AtoI               switch to calculate A to I editing (Default: off)
-tcf   --tcf-out            switch to write trimmed and collapsed fasta file (Default: off)
-gff   --gff-out            switch to output isomiR results in gff format (Default: off) 
-bam   --bam-out            switch to output results in bam format (Default: off) 
-trf   --tRNA-frag          switch to analyze tRNA fragment and halves (Default: off)
-o     --outDir             the directory of the outputs (Default: current directory) 
-dex   --diffex             perform differential expression with DESeq2 (Default: off)
-mdt   --metadata           the path to metadata file (Default: off, require '.csv' file format if -dex is opted)
-cms   --chunkmbs           chunk memory in megabytes per thread to use during bowtie alignment (Default: 256)
-spl   --save-pkl           save collapsed reads in binary format for later runs (Default: off)
-rr    --resume             resume from collapsed reads (Default: off)
-shh   --quiet              enable quiet/silent mode, only show warnings and errors (Default: off)
''')
    group.add_argument('-s','--samples', nargs='*', required=True, help=argparse.SUPPRESS)
    group.add_argument('-db', '--mir-DB', default='miRBase', required=True, help=argparse.SUPPRESS) 
    group.add_argument('-lib', '--libraries-path', required=True, help=argparse.SUPPRESS)
    group.add_argument('-on','--organism-name', required=True, help=argparse.SUPPRESS)
    group.add_argument('-ex', '--crThreshold', default='0.1', help=argparse.SUPPRESS)
    group.add_argument('-phr', '--phred64', type=int, default=33, help=argparse.SUPPRESS)
    group.add_argument('-spk', '--spikeIn', action='store_true', default=False, help=argparse.SUPPRESS)
    group.add_argument('-ie', '--isoform-entropy', action='store_true', default=False, help=argparse.SUPPRESS)
    group.add_argument('-cpu', '--threads', type=int, default='0', help=argparse.SUPPRESS)
    group.add_argument('-ai', '--AtoI', action='store_true', default=False, help=argparse.SUPPRESS)
    group.add_argument('-tcf', '--tcf-out', action='store_true', default=False, help=argparse.SUPPRESS)
    group.add_argument('-bam', '--bam-out', action='store_true', default=False, help=argparse.SUPPRESS)
    group.add_argument('-gff', '--gff-out', action='store_true', default=False, help=argparse.SUPPRESS)
    group.add_argument('-trf', '--tRNA-frag', action='store_true', default=False, help=argparse.SUPPRESS)
    group.add_argument('-o', '--outDir', help=argparse.SUPPRESS)
    group.add_argument('-dex', '--diffex', action='store_true', default=False, help=argparse.SUPPRESS)
    group.add_argument('-mdt', '--metadata', help=argparse.SUPPRESS)
    group.add_argument('-onam', '--outDirName', help=argparse.SUPPRESS)
    group.add_argument('-cms', '--chunkmbs', type=int, default=256, help=argparse.SUPPRESS)
    group.add_argument('-spl', '--save-pkl', action='store_true', default=False, help=argparse.SUPPRESS)
    group.add_argument('-rr', '--resume', action='store_true', default=False, help=argparse.SUPPRESS)
    group.add_argument('-shh',"--quiet", default=False, action='store_true', help=argparse.SUPPRESS)

    group1 = parser.add_argument_group("Data pre-processing", description='''-a,    --adapter            Sequence of a 3' adapter. The adapter and subsequent bases are trimmed
-g,    --front              Sequence of a 5' adapter. The adapter and any preceding bases are trimmed
-u,    --cut                Remove bases from each read. If LENGTH is positive, remove bases from the beginning. If LENGTH is negative, remove bases from the end
-nxt,  --nextseq-trim       NextSeq-specific quality trimming (each read). Trims also dark cycles appearing as high-quality G bases
-q,    --quality-cutoff     Trim low-quality bases from 5' and/or 3' ends of each read before adapter removal. If one value is given, only the 3' end is trimmed
                            If two comma-separated cutoffs are given, the 5' end is trimmed with the first cutoff, the 3' end with the second
-l,    --length             Shorten reads to LENGTH. Positive values remove bases at the end while negative ones remove bases at the beginning. This and the following
                            modifications are applied after adapter trimming
-NX,   --trim-n             Trim N's on ends of reads
-m,    --minimum-length     Discard reads shorter than LEN. (Default: 16)
-umi,  --uniq-mol-ids       Trim nucleotides of specific length at 5’ and 3’ ends of the read, after adapter trimming. eg: 4,4 or 0,4. (Use -udd to remove PCR duplicates)  
-udd,  --umiDedup           Specifies argument to removes PCR duplicates (Default: False); if TRUE it will remove UMI and remove PCR duplicates otherwise it only remove UMI and keep the raw counts (Require -umi option)
-qumi, --qiagenumi          Removes PCR duplicates of reads obtained from Qiagen platform (Default: Illumina; "-umi x,y " Required)


''')
    group1.add_argument("-a", "--adapter", type=lambda x: ("back", x), action="append",
        default=[], dest="adapters", help=argparse.SUPPRESS)
    group1.add_argument("-g", "--front", type=lambda x: ("front", x), action="append",
        default=[], dest="adapters", help=argparse.SUPPRESS)
    group1.add_argument("-u", "--cut", action='append', default=[], type=int, metavar="LENGTH", help=argparse.SUPPRESS)
    group1.add_argument("-nxt","--nextseq-trim", type=int, default=None, metavar="3'CUTOFF", help=argparse.SUPPRESS)
    group1.add_argument("-q", "--quality-cutoff", default="10", metavar="[5'CUTOFF,]3'CUTOFF", help=argparse.SUPPRESS)
    group1.add_argument("--length", "-l", type=int, default=3, metavar="LENGTH", help=argparse.SUPPRESS)
    group1.add_argument("-NX", "--trim-n", action='store_true', default=False, help=argparse.SUPPRESS)
    group1.add_argument("-m", "--minimum-length", default=16, help=argparse.SUPPRESS)
    group1.add_argument("-umi", "--uniq-mol-ids", default=None, help=argparse.SUPPRESS)
    group1.add_argument("-qumi", "--qiagenumi", action='store_true', default=False, help=argparse.SUPPRESS)
    group1.add_argument("-udd", "--umiDedup", action='store_true', default=False, help=argparse.SUPPRESS)
    #### we use none of the following cutadapt options but are required to pass default values for miRNA and cutadapt pipeline
    group1.add_argument("-op", "--output", metavar="FILE", help=argparse.SUPPRESS) #"Default: write to standard output"
    group1.add_argument("--compression-level", type=int, default=6, help=argparse.SUPPRESS)
    group1.add_argument("--overlap", type=int, metavar="MINLENGTH", default=3, help=argparse.SUPPRESS)
    group1.add_argument("--error-rate", type=float, default=0.12, metavar="RATE",help=argparse.SUPPRESS)
    group1.add_argument("--gc-content", type=float, default=50,  help=argparse.SUPPRESS)
    group1.add_argument("-cuv","--cutadaptVersion", help=argparse.SUPPRESS)
    group1.add_argument("-buv","--bowtieVersion", help=argparse.SUPPRESS)
    group1.add_argument("--action", choices=('trim', 'mask', 'lowercase', 'none'), default='trim', help=argparse.SUPPRESS)
    group1.add_argument("-n", "--times", type=int, metavar="COUNT", default=1, help=argparse.SUPPRESS)
    group1.add_argument("--match-read-wildcards", action="store_true", default=False,help=argparse.SUPPRESS)
    group1.add_argument("-N", "--no-match-adapter-wildcards", action="store_false", default=True, dest='match_adapter_wildcards',help=argparse.SUPPRESS)
    group1.add_argument("--fasta", default=False, action='store_true',help=argparse.SUPPRESS)
    group1.add_argument("--buffer-size", type=int, default=4000000, help=argparse.SUPPRESS)
    group1.add_argument("--no-indels", action='store_false', dest='indels', default=True, help=argparse.SUPPRESS)
    group1.add_argument("-M", "--maximum-length", default=None, type=int, metavar="LEN[:LEN2]", help=argparse.SUPPRESS)
    group1.add_argument("--numba-pll", default=None, action='store_false', help=argparse.SUPPRESS)
    group1.add_argument("--numba-cuda", default=None, action='store_false', help=argparse.SUPPRESS)
    
    group4 = parser.add_argument_group('miRNA Error Correction', description='''microRNA correction method for single base substitutions due to sequencing errors (Note: Refines reads at the expense of time)
-mEC,  --miREC              Enable miRNA error correction (miREC)
-kh,   --threshold          the value for frequency threshold τ (Default kh = 5)
-ks,   --kmer-start         kmer range start value (k_1, default 15) 
-ke,   --kmer-end           kmer range end value (k_end, default 20)
''')
    group4.add_argument('-mEC','--miREC', help=argparse.SUPPRESS, action='store_true')
    group4.add_argument('-kh','--threshold', help=argparse.SUPPRESS, default='5', metavar='', type=int)
    group4.add_argument('-ks','--kmer-start', help=argparse.SUPPRESS, default='15', metavar='', type=int)
    group4.add_argument('-ke','--kmer-end', help=argparse.SUPPRESS, default='20', metavar='', type=int)
    ## group - 2 ##

    group2 = parser.add_argument_group('Predicting novel miRNAs',description='''The predictive model for novel miRNA detection is trained on human and mouse!
-nmir, --novel-miRNA        include prediction of novel miRNAs
-minl, --minLength          the minimum length of the retained reads for novel miRNA detection (default: 16)
-maxl, --maxLength          the maximum length of the retained reads for novel miRNA detection (default: 25)
-c,    --minReadCounts      the minimum read counts supporting novel miRNA detection (default: 2)
-mloc, --maxMappingLoci     the maximum number of mapping loci for the retained reads for novel miRNA detection (default: 3)
-sl,   --seedLength         the seed length when invoking Bowtie for novel miRNA detection (default: 25)
-olc,  --overlapLenCutoff   the length of overlapped seqence when joining reads into longer sequences based on the coordinate 
                            on the genome for novel miRNA detection (default: 14)
-clc,  --clusterLength      the maximum length of the clustered sequences for novel miRNA detection (default: 30)
''')
    group2.add_argument('-nmir', '--novel-miRNA', help=argparse.SUPPRESS, action='store_true')
    group2.add_argument('-minl', '--minLength', type=int, default='16', metavar='', help=argparse.SUPPRESS)
    group2.add_argument('-maxl', '--maxLength', type=int, default='25', metavar='', help=argparse.SUPPRESS)
    group2.add_argument('-c', '--minReadCounts', default='2', metavar='', help=argparse.SUPPRESS)
    group2.add_argument('-mloc', '--maxMappingLoci', default='3', metavar='', help=argparse.SUPPRESS)
    group2.add_argument('-sl','--seedLength', default='25', metavar='', help=argparse.SUPPRESS)
    group2.add_argument('-olc', default='14', dest='overlapLenCutoff', metavar='', help=argparse.SUPPRESS)
    group2.add_argument('-clc', default='30', dest='clusterLength', metavar='', help=argparse.SUPPRESS)


    ## group - 3 ##

    group3 = parser.add_argument_group('Optional PATH arguments',
	description='''-pbwt, --bowtie-path        the path to system's directory containing bowtie binary
-psam, --samtools-path      the path to system's directory containing samtools binary
-prf,  --RNAfold-path       the path to system's directory containing RNAfold binary
''')
    group3.add_argument('-pbwt', '--bowtie-path', metavar="", help=argparse.SUPPRESS)
    group3.add_argument('-psam', '--samtools-path', metavar="", help=argparse.SUPPRESS)
    group3.add_argument('-prf', '--RNAfold-path', metavar="", help=argparse.SUPPRESS)
    
    return parser.parse_args()

    

#group.add_argument('-ad', default='none', dest='adapter', metavar='<string>', help='the adapter need to be removed which could be illumina, ion or a defined sequence (default: none)')



from Bio.Seq import Seq
from Bio import SeqIO
from Bio import pairwise2
import sys

def calculate_identity(sequenceA, sequenceB):
	#sequenceA is the clustered seq and sequenceB is the read
	#Returns the percentage of identical characters between two sequences.
	sa, sb, sl = sequenceA, sequenceB, len(sequenceA)
	matches = 0
	for i in range(sl):
		if sa[i] == sb[i] and sa[i] != '-' and  sb[i] != '-':
			matches = matches + 1
	return matches

def chunks(seq, n):
	return (seq[i:i+n] for i in range(0, len(seq), n))

class ReadCluster(object):
	threshold = 0.8
	headShiftStep = 3
	tailShiftStep = 6
	def __init__(self, chrSeqDic, chr, strand, startPos, endPos, clusterName, clusterSeq, readNameList, readSeqList, readCountList):
		self.chrSeqDic = chrSeqDic
		self.chr = chr
		self.strand = strand
		self.startPos = startPos
		self.endPos = endPos
		self.clusterName = clusterName
		self.clusterSeq = clusterSeq
		self.readNameList = readNameList
		self.readSeqList = readSeqList
		self.readCountList = readCountList

	def totalreadCount(self):
		return sum(self.readCountList)

	def totalseqCount(self):
		return len(self.readSeqList)

	def align2Standard(self):
		#align = Alignment(Gapped(IUPAC.unambiguous_dna, "-"))
		alignSeqList = []
		clusterSeqTmp = self.clusterSeq
		exactMatchCount = 0
		#for seq in self.readSeqList:
		for i in range(len(self.readSeqList)):
			# Perform pairwise local alignment. And increase the penalty of gap opening and extending into -10 and -10,
			# so that there will be no gap in the alignned result.
			seq = self.readSeqList[i]
			alignTmp = pairwise2.align.localms(clusterSeqTmp, seq, 2, -1, -20, -20)
			clusterSeqTmpNew = alignTmp[0][0]
			seqTmp = alignTmp[0][1]
			identity = calculate_identity(clusterSeqTmpNew, seqTmp)
			if identity == len(seq):
				exactMatchCount = exactMatchCount + self.readCountList[i]
			else:
				pass
			if len(alignSeqList) == 0:
				alignSeqList.append(clusterSeqTmpNew)
				alignSeqList.append(seqTmp)
			else:
				if clusterSeqTmpNew == alignSeqList[0]:
					alignSeqList.append(seqTmp)
				else:
					headAdd1 = clusterSeqTmpNew.index(clusterSeqTmp)
					tailAdd1 = len(clusterSeqTmpNew)-headAdd1-len(clusterSeqTmp)
					headAdd2 = alignSeqList[0].index(clusterSeqTmp)
					tailAdd2 = len(alignSeqList[0])-headAdd2-len(clusterSeqTmp)
					if headAdd1 >= headAdd2 and tailAdd1 >= tailAdd2:
						for j in range(len(alignSeqList)):
							alignSeqList[j] = (headAdd1 - headAdd2)*'-'+alignSeqList[j]+(tailAdd1 - tailAdd2)*'-'
						alignSeqList.append(seqTmp)
					elif headAdd1 >= headAdd2 and tailAdd1 < tailAdd2:
						for j in range(len(alignSeqList)):
							alignSeqList[j] = (headAdd1 - headAdd2)*'-'+alignSeqList[j]
						alignSeqList.append(seqTmp+(tailAdd2 - tailAdd1)*'-')
					elif headAdd1 < headAdd2 and tailAdd1 >= tailAdd2:
						for j in range(len(alignSeqList)):
							alignSeqList[j] = alignSeqList[j]+(tailAdd1 - tailAdd2)*'-'
						alignSeqList.append((headAdd2 - headAdd1)*'-'+seqTmp)
					elif headAdd1 < headAdd2 and tailAdd1 < tailAdd2:
						alignSeqList.append((headAdd2 - headAdd1)*'-'+seqTmp+(tailAdd2 - tailAdd1)*'-')
					else:
						pass
		return (alignSeqList, exactMatchCount)
	
	def locateStartPosition(self):
		alignSeqList1, exactMatchCount1 = self.align2Standard()
		clusterSeq = alignSeqList1[0]
		transformedListTmp = []
		for i in range(len(clusterSeq)):
			for j in range(1, len(alignSeqList1)):
				transformedListTmp.append(alignSeqList1[j][i])
		transformedList = list(chunks(transformedListTmp, len(alignSeqList1)-1))
		readCountSum = sum(self.readCountList)
		clusterSeqRatioList = []
		majorSeqRationList = []
		majorSeq = ""
		for i in range(len(transformedList)):
			item = transformedList[i]
			count_A = 0
			count_T = 0
			count_C = 0
			count_G = 0
			for j in range(len(item)):
				if item[j] == 'A':
					count_A = count_A + self.readCountList[j]
				elif item[j] == 'T':
					count_T = count_T + self.readCountList[j]
				elif item[j] == 'C':
					count_C = count_C + self.readCountList[j]
				elif item[j] == 'G':
					count_G = count_G + self.readCountList[j]
				else:
					pass
			if clusterSeq[i] == 'A':
				clusterSeqPositionRatio = float(count_A)/readCountSum
			elif clusterSeq[i] == 'T':
				clusterSeqPositionRatio = float(count_T)/readCountSum
			elif clusterSeq[i] == 'C':
				clusterSeqPositionRatio = float(count_C)/readCountSum
			elif clusterSeq[i] == 'G':
				clusterSeqPositionRatio = float(count_G)/readCountSum
			else:
				clusterSeqPositionRatio = 0
			clusterSeqRatioList.append(clusterSeqPositionRatio)
			listTmp = [[count_A, 'A'], [count_T, 'T'], [count_C, 'C'], [count_G, 'G']]
			listTmp.sort(reverse=True)
			majorSeqRationList.append(float(listTmp[0][0])/readCountSum)
			majorSeq = majorSeq + listTmp[0][1]
		headStartPosition = None
		tailStartPosition = None
		for i in range(len(majorSeqRationList)):
			if float(majorSeqRationList[i]) >= ReadCluster.threshold:
				headStartPosition = i
				break
		for i in range(-1, -len(majorSeqRationList)-1, -1):
			if float(majorSeqRationList[i]) >= ReadCluster.threshold:
				tailStartPosition = i
				break
		return (alignSeqList1, exactMatchCount1, clusterSeq, clusterSeqRatioList, majorSeq, majorSeqRationList, headStartPosition, tailStartPosition)

	def calculateFeature(self):
		alignSeqList2, exactMatchCount2, clusterSeq, clusterSeqRatioList, majorSeq, majorSeqRationList, headStartPosition, tailStartPosition = self.locateStartPosition()
		#alignSeqList2, exactMatchCount2 = self.align2Standard()
		#for subseq in alignSeqList2:
		#	print subseq
		#print ReadCluster.headShiftStep
		#print ReadCluster.tailShiftStep
		neclueotidePositionList = []
		for i in range(ReadCluster.headShiftStep+ReadCluster.tailShiftStep):
			if i <= ReadCluster.headShiftStep-1:
				postionLable = 'head_minus%d_'%(ReadCluster.headShiftStep-i)
			else:
				postionLable = 'tail_plus%d_'%(i+1-ReadCluster.headShiftStep)
			for neclueotide in ['templateNucleotide', 'TemplateNucleotide_percentage', 'nonTemplateNucleotide_percentage', 'A_percentage', 'T_percentage', 'C_percentage', 'G_percentage']:
				neclueotidePositionList.append(postionLable+neclueotide)

		if (headStartPosition is not None )and (tailStartPosition is not None):
			readCountSum = sum(self.readCountList)
			seqCount = len(self.readSeqList)
			exactMatchRatio = float(exactMatchCount2)/readCountSum
			headAdd = headStartPosition
			tailAdd = -tailStartPosition-1
			headUnstableLength = headAdd
			tailUnstableLength = tailAdd
			#print tailAdd
			if headAdd < ReadCluster.headShiftStep:
				for i in range(len(alignSeqList2)):
					alignSeqList2[i] = (ReadCluster.headShiftStep-headAdd)*'-'+alignSeqList2[i]
				headStartPosition = ReadCluster.headShiftStep
			else:
				pass
			if tailAdd < ReadCluster.tailShiftStep:
				for i in range(len(alignSeqList2)):
					alignSeqList2[i] = alignSeqList2[i]+(ReadCluster.tailShiftStep-tailAdd)*'-'
				tailStartPosition = 0 - ReadCluster.tailShiftStep
			else:
				pass
			adjustedClusterSeq = alignSeqList2[0]
			headDashCount = 0
			for i in alignSeqList2[0]:
				if i == '-':
					headDashCount += 1
				else:
					break
			tailDashCount = 0
			for i in reversed(alignSeqList2[0]):
				if i == '-':
					tailDashCount += 1
				else:
					break

			if self.strand == '+':
				genomeSeqStart = self.startPos - headDashCount
				genomeSeqEnd = self.endPos + tailDashCount
				tmp_new = self.chrSeqDic[self.chr][genomeSeqStart-1:genomeSeqEnd]
				templateSeq = tmp_new.upper()
			else:
				genomeSeqStart = self.startPos - tailDashCount
				genomeSeqEnd = self.endPos + headDashCount
				tmp_new = self.chrSeqDic[self.chr][genomeSeqStart-1:genomeSeqEnd]
				templateSeq = str(Seq(tmp_new.upper()).reverse_complement())
			alignSeqList2.insert(0, templateSeq)
			clusterSecondSeq = alignSeqList2[2]

			colList = list(range(headStartPosition-ReadCluster.headShiftStep, headStartPosition)) + list(range(tailStartPosition, tailStartPosition+ReadCluster.tailShiftStep))
			transformedListTmp = []
			for i in range(len(colList)):
				for j in range(2, len(alignSeqList2)):
					transformedListTmp.append(alignSeqList2[j][colList[i]])
			transformedList = list(chunks(transformedListTmp, len(alignSeqList2)-2))
			neclueotideCountList = []
			#print len(self.readCountList)
			#print len(transformedList)
			for i in range(len(transformedList)):
				try:
					templateNucleotide = templateSeq[colList[i]]
				except IndexError:
					print("Error happends at: %s"%(self.clusterName))
					print("Please check it.")
					print("templateSeq is: %s"%(templateSeq))
					print("index is:")
					print(colList[i])
					sys.exit(1)
				item = transformedList[i]
				#print len(item)
				count_A = 0
				count_T = 0
				count_C = 0
				count_G = 0
				count_Dash = 0
				for j in range(len(item)):
					if item[j] == 'A':
						count_A = count_A + self.readCountList[j]
					elif item[j] == 'T':
						count_T = count_T + self.readCountList[j]
					elif item[j] == 'C':
						count_C = count_C + self.readCountList[j]
					elif item[j] == 'G':
						count_G = count_G + self.readCountList[j]
					else:
						count_Dash = count_Dash + self.readCountList[j]

				if templateNucleotide == 'A':
					count_templateNucleotide = count_A
					count_nonTemplateNucleotide = count_T+count_C+count_G
				elif templateNucleotide == 'T':
					count_templateNucleotide = count_T
					count_nonTemplateNucleotide = count_A+count_C+count_G
				elif templateNucleotide == 'C':
					count_templateNucleotide = count_C
					count_nonTemplateNucleotide = count_A+count_T+count_G
				elif templateNucleotide == 'G':
					count_templateNucleotide = count_G
					count_nonTemplateNucleotide = count_A+count_T+count_C
				else:
					count_templateNucleotide = 0
					count_nonTemplateNucleotide = 0

				if count_A+count_T+count_C+count_G != 0:
					count_TemplateNucleotide_percentage = float(count_templateNucleotide)/(count_A+count_T+count_C+count_G)
					count_nonTemplateNucleotide_percentage = float(count_nonTemplateNucleotide)/(count_A+count_T+count_C+count_G)
					if templateNucleotide == 'A':
						if count_nonTemplateNucleotide != 0:
							count_A_percentage = 0
							count_T_percentage = float(count_T)/count_nonTemplateNucleotide
							count_C_percentage = float(count_C)/count_nonTemplateNucleotide
							count_G_percentage = float(count_G)/count_nonTemplateNucleotide
						else:
							count_A_percentage = 0
							count_T_percentage = 0
							count_C_percentage = 0
							count_G_percentage = 0
					elif templateNucleotide == 'T':
						if count_nonTemplateNucleotide != 0:
							count_A_percentage = float(count_A)/count_nonTemplateNucleotide
							count_T_percentage = 0
							count_C_percentage = float(count_C)/count_nonTemplateNucleotide
							count_G_percentage = float(count_G)/count_nonTemplateNucleotide
						else:
							count_A_percentage = 0
							count_T_percentage = 0
							count_C_percentage = 0
							count_G_percentage = 0
					elif templateNucleotide == 'C':
						if count_nonTemplateNucleotide != 0:
							count_A_percentage = float(count_A)/count_nonTemplateNucleotide
							count_T_percentage = float(count_T)/count_nonTemplateNucleotide
							count_C_percentage = 0
							count_G_percentage = float(count_G)/count_nonTemplateNucleotide
						else:
							count_A_percentage = 0
							count_T_percentage = 0
							count_C_percentage = 0
							count_G_percentage = 0
					elif templateNucleotide == 'G':
						if count_nonTemplateNucleotide != 0:
							count_A_percentage = float(count_A)/count_nonTemplateNucleotide
							count_T_percentage = float(count_T)/count_nonTemplateNucleotide
							count_C_percentage = float(count_C)/count_nonTemplateNucleotide
							count_G_percentage = 0
						else:
							count_A_percentage = 0
							count_T_percentage = 0
							count_C_percentage = 0
							count_G_percentage = 0
					else:
						pass
					
				else:
					count_TemplateNucleotide_percentage = 0
					count_nonTemplateNucleotide_percentage = 0
					count_A_percentage = 0
					count_T_percentage = 0
					count_C_percentage = 0
					count_G_percentage = 0

				neclueotideCountList = neclueotideCountList + [templateNucleotide, count_TemplateNucleotide_percentage, count_nonTemplateNucleotide_percentage, count_A_percentage, count_T_percentage, count_C_percentage, count_G_percentage]
				
			outList = [self.clusterName, self.clusterSeq, adjustedClusterSeq, clusterSecondSeq, templateSeq, seqCount, readCountSum, exactMatchRatio, headUnstableLength, tailUnstableLength, neclueotidePositionList, neclueotideCountList, True]
		else:
			outList = [self.clusterName, self.clusterSeq, None, None, None, None, None, None, None, None, neclueotidePositionList, None, False]
		return outList

	def calculateFeature_old_version(self):
		alignSeqList2, exactMatchCount2, clusterSeq, clusterSeqRatioList, majorSeq, majorSeqRationList, headStartPosition, tailStartPosition = self.locateStartPosition()
		#alignSeqList2, exactMatchCount2 = self.align2Standard()
		#for subseq in alignSeqList2:
		#	print subseq
		#print ReadCluster.headShiftStep
		#print ReadCluster.tailShiftStep
		readCountSum = sum(self.readCountList)
		seqCount = len(self.readSeqList)
		exactMatchRatio = float(exactMatchCount2)/readCountSum
		headAdd = alignSeqList2[0].index(self.clusterSeq)
		tailAdd = len(alignSeqList2[0])-headAdd-len(self.clusterSeq)
		#print tailAdd
		if headAdd < ReadCluster.headShiftStep:
			for i in range(len(alignSeqList2)):
				alignSeqList2[i] = (ReadCluster.headShiftStep-headAdd)*'-'+alignSeqList2[i]
		else:
			pass
		if tailAdd < ReadCluster.tailShiftStep:
			for i in range(len(alignSeqList2)):
				alignSeqList2[i] = alignSeqList2[i]+(ReadCluster.tailShiftStep-tailAdd)*'-'
		else:
			pass
		start = alignSeqList2[0].index(self.clusterSeq)
		end = alignSeqList2[0].index(self.clusterSeq)+len(self.clusterSeq)-1
		colList = list(range(start-ReadCluster.headShiftStep, start+ReadCluster.headShiftStep+1)) + list(range(end-ReadCluster.tailShiftStep, end+ReadCluster.tailShiftStep+1))
		transformedListTmp = []
		for i in range(len(colList)):
			for j in range(1, len(alignSeqList2)):
				transformedListTmp.append(alignSeqList2[j][colList[i]])
		transformedList = list(chunks(transformedListTmp, len(alignSeqList2)-1))
		neclueotideCountList = []
		neclueotidePositionList = []
		for i in range(len(transformedList)):
			item = transformedList[i]
			count_A = 0
			count_T = 0
			count_C = 0
			count_G = 0
			for j in range(len(item)):
				if item[j] == 'A':
					count_A = count_A + self.readCountList[j]
				elif item[j] == 'T':
					count_T = count_T + self.readCountList[j]
				elif item[j] == 'C':
					count_C = count_C + self.readCountList[j]
				elif item[j] == 'G':
					count_G = count_G + self.readCountList[j]
				else:
					pass
			neclueotideCountList = neclueotideCountList + [count_A,count_T,count_C,count_G]
			#neclueotideCountList.append(count_A,count_T,count_C,count_G)
			if i <= ReadCluster.headShiftStep-1:
				postionLable = 'head_minus%d_'%(ReadCluster.headShiftStep-i)
			elif i == ReadCluster.headShiftStep:
				postionLable = 'head_0_'
			elif i > ReadCluster.headShiftStep and i <= ReadCluster.headShiftStep*2:
				postionLable = 'head_plus%d_'%(i-ReadCluster.headShiftStep)
			elif i > ReadCluster.headShiftStep*2 and i < ReadCluster.headShiftStep*2+ReadCluster.tailShiftStep+1:
				postionLable = 'tail_minus%d_'%(ReadCluster.headShiftStep*2+ReadCluster.tailShiftStep+1-i)
			elif i == ReadCluster.headShiftStep*2+ReadCluster.tailShiftStep+1:
				postionLable = 'tail_0_'
			elif i > ReadCluster.headShiftStep*2+ReadCluster.tailShiftStep+1:
				postionLable = 'tail_plus%d_'%(i-ReadCluster.headShiftStep*2-ReadCluster.tailShiftStep-1)
			else:
				pass
			for neclueotide in ['A', 'T', 'C', 'G']:
				neclueotidePositionList.append(postionLable+neclueotide)
		return (self.clusterName, self.clusterSeq, seqCount, readCountSum, exactMatchRatio, neclueotidePositionList, neclueotideCountList)



import mirge.forgi.graph.bulge_graph as fgb
import os
import re

def isisParenthese(i):
	flag = False
	if i == '(' or i == ')':
		flag = True
	return flag

#get miRNA position on RNASequence, 0-based
def getMiRNAPosition(miRNASeq, RNASequence):
	startPos = RNASequence.find(miRNASeq)
	endPos = startPos + len(miRNASeq)
	return [startPos, endPos]

#get miRNA structure
def getMiRNAStructure(miRNASeq, RNASequence, RNAStructure):
	[startPos, endPos] = getMiRNAPosition(miRNASeq, RNASequence)
	mirnaStruct = RNAStructure[startPos:endPos]
	return mirnaStruct

#get GC content of a sequence
def getGCContent(sequence):
	gc = sequence.count('G')
	gc = gc + sequence.count('g')
	gc = gc + sequence.count('C')
	gc = gc + sequence.count('c')
	GCContent = float(gc)/len(sequence)
	return GCContent

#check whether there is a loop in mature miRNA
def getMirnaIncludedInLoop( RNASequence, RNAStructure, miRNASeq):
	flag = False
	mirnaStruct = getMiRNAStructure(miRNASeq, RNASequence, RNAStructure)
	#bg2 = fgb.BulgeGraph()
	#fa2 = '\n'.join(['>miRNAStr', miRNASeq, mirnaStruct])
	#bg2.from_fasta(fa2)
	#for i in range(1, len(miRNASeq)+1):
	#	if bg2.pairing_partner(i)
	pattern = re.compile("\(.*\)")
	pattern.search(mirnaStruct)
	if(('(' in mirnaStruct) and (')' in mirnaStruct)):
		if pattern.search(mirnaStruct) is None:
			pass
		else:
			flag = True
	return flag

#check miRNA located in which arm
def checkArm(RNASequence, RNAStructure, miRNASeq):
	armDetailedType = ''
	if(getMirnaIncludedInLoop(RNASequence, RNAStructure, miRNASeq)):
		armDetailedType = 'loop'
	#check arms
	else:
		mirnaStruct = getMiRNAStructure(miRNASeq, RNASequence, RNAStructure)
		if '(' in mirnaStruct :
			if mirnaStruct.count('(') >= mirnaStruct.count(')'):
				armDetailedType = 'arm5'
			else:
				armDetailedType = 'arm3'
		elif ')' in mirnaStruct :
			if mirnaStruct.count(')') >= mirnaStruct.count('('):
				armDetailedType = 'arm3'
			else:
				armDetailedType = 'arm5'
		else:
			armDetailedType = 'unmatchedRegion'
	return armDetailedType
#end fun

def decorate_str(RNAStructue):
	open_brackets1 = '('
	close_brackets1 = ')'
	brackets_map1 = {')': '('}
	stack1 = []
	errorFrontPosition1 = []
	for i, char in enumerate(RNAStructue):
		if char in open_brackets1:
			stack1.append(char)
		elif char in close_brackets1:
			if len(stack1) < 1:
				stack1.append(char)
				errorFrontPosition1.append(i)
			elif brackets_map1[char] == stack1[-1]:
				stack1.pop()
			else:
				stack1.append(char)
				errorFrontPosition1.append(i)
		else:
			continue
	open_brackets2 = ')'
	close_brackets2 = '('
	brackets_map2 = {'(': ')'}
	stack2 = []
	errorFrontPosition2 = []
	for i in range(len(RNAStructue)-1, -1, -1):
		char = RNAStructue[i]
		if char in open_brackets2:
			stack2.append(char)
		elif char in close_brackets2:
			if len(stack2) < 1:
				stack2.append(char)
				errorFrontPosition2.append(i)
			elif brackets_map2[char] == stack2[-1]:
				stack2.pop()
			else:
				stack2.append(char)
				errorFrontPosition2.append(i)
		else:
			continue
	errorFrontPosition = errorFrontPosition1+errorFrontPosition2
	decorated_RNAStr = ''
	for i, char in enumerate(RNAStructue):
		if i in errorFrontPosition:
			decorated_RNAStr = decorated_RNAStr + '.'
		else:
			decorated_RNAStr = decorated_RNAStr + char
	return decorated_RNAStr

# Statisitcally analyze the list such as: ['s0', 's0', 'i0', 'i0','s1', 's1'...], location is 0-based
def statisList(elementIndexList, RNASeq):
# Here , the length of elementIndexList and RNASeq must be equal.
	elementSeq = []
	elementDic = {}
	i = 0
	for item in elementIndexList:
		if item not in elementSeq:
			elementSeq.append(item)
			elementDic.update({item:[]})
		i = i + 1

# Rewrote function of 'to_element_string', because the lower version forgi 0.2 lack the option of with_numbers=True
def to_element_string(bulgeGraphInstance, with_numbers=False):
	output_str = [' '] * (bulgeGraphInstance.seq_length + 1)
	output_nr = [' '] * (bulgeGraphInstance.seq_length + 1)
	for d in bulgeGraphInstance.defines.keys():
		for resi in bulgeGraphInstance.define_residue_num_iterator(d, adjacent=False):
			output_str[resi] = d[0]
			output_nr[resi] = d[-1]
	if with_numbers:
		return "".join(output_str).strip()+"\n"+"".join(output_nr).strip()
	else:
		return "".join(output_str).strip()

def chunks(arr, n):
	return [arr[i:i+n] for i in range(0, len(arr), n)]

class ReadPrecusor(object):
	def __init__(self, rnafoldCmdTmp, premiRNASeqName, premiRNASeq, premiRNAStructure, mfe, tmpdir, prunedLength, *miRNASeq):
		# The first element in the list of *miRNASeq must be the miRNAs in the pre-MiRNA.
		# The features about the miRNA should be about the first element.
		# If *miRNASeq has two elements, the second one will be the other one which is only used to generate the corePrecusor.
		self.rnafoldCmdTmp = rnafoldCmdTmp
		self.premiRNASeqName = premiRNASeqName
		self.premiRNASeq = premiRNASeq
		#self.direction = premiRNASeq
		self.premiRNAStructure = premiRNAStructure
		self.mfe = mfe
		self.tmpdir = tmpdir
		self.miRNASeq = miRNASeq
		self.prunedLength = int(prunedLength)
		#Here, self.miRNASeq is a tuple which stores one or two mirNA sequences.

	def getPrunedPrecusor(self):
		if len(self.miRNASeq) == 1:
			miRNASeq1 = self.miRNASeq[0]
			selectedMiRNASeq = miRNASeq1
		else:
			miRNASeq1 = self.miRNASeq[0]
			miRNASeq2 = self.miRNASeq[1]
			startPos1, endPos1 = getMiRNAPosition(miRNASeq1, self.premiRNASeq)
			if startPos1 == 20:
				selectedMiRNASeq = miRNASeq1
			else:
				selectedMiRNASeq = miRNASeq2
		armType = checkArm(self.premiRNASeq, self.premiRNAStructure, selectedMiRNASeq)
		bg1 = fgb.BulgeGraph()
		fa = '\n'.join(['>PrecusorStr', self.premiRNASeq, self.premiRNAStructure])
		bg1.from_fasta(fa)
		if armType == 'arm5':
			startPos1, endPos1 = getMiRNAPosition(selectedMiRNASeq, self.premiRNASeq)
			otherCount = 0 
			for i in range(startPos1, endPos1):
				if self.premiRNAStructure[i] == '(':
					break
				otherCount = otherCount + 1
			# Locate the partner of a base pair
			try:
				pairedPos = bg1.pairing_partner(startPos1+otherCount+1) - 1
				if otherCount + pairedPos >= len(self.premiRNASeq):
					newEndPos = len(self.premiRNASeq) - 1
				else:
					newEndPos = otherCount + pairedPos
				prunedPrecursorSeq = self.premiRNASeq[startPos1-self.prunedLength : newEndPos+1+self.prunedLength]
				prunedPrecusorStr = self.premiRNAStructure[startPos1-self.prunedLength : newEndPos+1+self.prunedLength]
			except TypeError:
				prunedPrecursorSeq = None
				prunedPrecusorStr = None
		else:
			startPos1, endPos1 = getMiRNAPosition(selectedMiRNASeq, self.premiRNASeq)
			otherCount = 0 
			for i in range(endPos1-1, -1, -1):
				if self.premiRNAStructure[i] == ')':
					break
				otherCount = otherCount + 1
			# Locate the partner of a base pair
			try:
				pairedPos = bg1.pairing_partner(endPos1-otherCount) - 1
				if pairedPos - otherCount < 0:
					newStartPos = 0
				else:
					newStartPos = pairedPos - otherCount
				prunedPrecursorSeq = self.premiRNASeq[newStartPos-self.prunedLength : endPos1+self.prunedLength]
				prunedPrecusorStr = self.premiRNAStructure[newStartPos-self.prunedLength : endPos1+self.prunedLength]
			except TypeError:
				prunedPrecursorSeq = None
				prunedPrecusorStr = None
		# Decorate the CorePrecusor structure.
		if not (prunedPrecusorStr is None):
			prunedPrecusorStr = decorate_str(prunedPrecusorStr)
			
		# Predict the new Structure for the prunedPrecursorSeq
		if not (prunedPrecursorSeq is None):
			with open(os.path.join(self.tmpdir, 'prunedPrecusor.fa'), 'w') as outf:
				fa_tmp = '\n'.join(['>'+self.premiRNASeqName, prunedPrecursorSeq])
				outf.write(fa_tmp+'\n')
			f1 = os.path.join(self.tmpdir, 'prunedPrecusor.fa')
			f2 = os.path.join(self.tmpdir, 'prunedPrecusor.str')
			os.system('%s < %s --noPS --noLP > %s'%(self.rnafoldCmdTmp, f1, f2))
			try:	
				with open(f2, 'r') as inf:
					try:
						tmp = inf.readlines()
						prunedPrecusorStrNew = tmp[2].strip().split()[0]
						if len(tmp[2].strip().split()[1]) >= 3:
							mFE = float(tmp[2].strip().split()[1][1:-1].strip())
						else:
							mFE = float(tmp[2].strip().split()[2][:-1].strip())
					except:
						mFE = 0
						prunedPrecusorStrNew = prunedPrecusorStr
			except FileNotFoundError:
				prunedPrecusorStrNew = None
			os.remove(f1)
			os.remove(f2)
		else:
			prunedPrecusorStrNew = None
			mFE = None
		return (prunedPrecursorSeq, prunedPrecusorStrNew, mFE)

	def featuresInPrecusor(self):
		if not (self.premiRNAStructure  is None) and '(' in self.premiRNAStructure  and ')' in self.premiRNAStructure:
			bg = fgb.BulgeGraph()
			fa = '\n'.join(['>prunedPrecursorStr', self.premiRNASeq, self.premiRNAStructure])
			bg.from_fasta(fa)
			bindingCount1 = 0
			bindingCount2 = 0
			for item in self.premiRNAStructure:
				if item == '(' :
					bindingCount1 = bindingCount1 + 1
				if item == ')' :
					bindingCount2 = bindingCount2 + 1
			# Calculate bindingCount in the CorePrecusor
			bindingCount = max([bindingCount1, bindingCount2])
			# Calculate the hairpinCount and interiorLoopCount
			hairpinCount = 0
			interiorLoopCount = 0
			for string in bg.to_bg_string().strip().split('\n'):
				if 'define' in string:
					contentTmp = string.split(' ')
					elementType = contentTmp[1]
					if elementType[0] == 'h':
						hairpinCount = hairpinCount + 1
					if elementType[0] == 'i':
						interiorLoopCount = interiorLoopCount + 1
			# Calculate apicalLoopSize of the CorePrecusor
			# To get compatible with the older version of forgi, the below statement is abandoned.
			#elementTypeSeq, indeSeq = bg.to_element_string(with_numbers=True).split('\n')
			elementTypeSeq, indeSeq = to_element_string(bg, with_numbers=True).split('\n')
			
			elementIndexList = [elementTypeSeq[i]+indeSeq[i] for i in range(len(elementTypeSeq))]
			elementSeqList = []
			elementlocationList = []
			# the position is 0-based
			for index, item in enumerate(elementIndexList):
				if len(elementlocationList) == 0:
					elementSeqList.append(item)
					elementlocationList.append([index, index])
				elif item != elementSeqList[-1]:
					elementSeqList.append(item)
					elementlocationList.append([index, index])
				else:
					elementlocationList[-1][1] = index
			elementDic = {}
			for i in range(len(elementSeqList)):
				if elementSeqList[i] not in elementDic.keys():
					elementDic.update({elementSeqList[i] : [elementlocationList[i]]})
				else:
					elementDic[elementSeqList[i]].append(elementlocationList[i])
			hairpinRagion = []
			for d in elementDic.keys():
				if d[0] == 'h':
					hairpinRagion = hairpinRagion + elementDic[d]
			if len(hairpinRagion) >= 1:
				apicalLoopSize = max([len(range(item[0], item[1]+1)) for item in hairpinRagion])
			else:
				apicalLoopSize = 0
			armTypeOverlenDic = {}
			# Calculate the armType of the miRNA
			armType = checkArm(self.premiRNASeq, self.premiRNAStructure, self.miRNASeq[0])
			# Calculate the overlap of miRNA with the hairpin
			# startPos and endPos are the locations of the miRNA in the corePrecusor. 0-based
			startPos, endPos = getMiRNAPosition(self.miRNASeq[0], self.premiRNASeq)
			endPos = endPos - 1
			if len(hairpinRagion) >= 1:
				overlapLen = max([len(set(range(startPos, endPos+1)) & set(range(item[0], item[1]+1))) for item in hairpinRagion])
			else:
				overlapLen = 0
			if overlapLen == 0:
				if armType == 'arm5':
					distanceToLoop = min([item[0]-1-endPos for item in hairpinRagion])
				else:
					distanceToLoop = min([startPos-item[1]-1 for item in hairpinRagion])
			else:
				for item in hairpinRagion:
					if len(set(range(startPos, endPos+1)) & set(range(item[0], item[1]+1))) == overlapLen:
						if armType == 'arm5':
							distanceToLoop = item[0]-1-endPos
						else:
							distanceToLoop = startPos - item[1] - 1
						break
			armTypeOverlenDic.update({armType: [overlapLen, distanceToLoop]})
			if len(self.miRNASeq) == 2:
				armType2 = checkArm(self.premiRNASeq, self.premiRNAStructure, self.miRNASeq[1])
				startPos2, endPos2 = getMiRNAPosition(self.miRNASeq[1], self.premiRNASeq)
				endPos2 = endPos2 - 1
				if len(hairpinRagion) >= 1:
					overlapLen2 = max([len(set(range(startPos2, endPos2+1)) & set(range(item[0], item[1]+1))) for item in hairpinRagion])
				else:
					overlapLen2 = 0
				if overlapLen2 == 0:
					if armType2 == 'arm5':
						distanceToLoop2 = min([item[0]-1-endPos2 for item in hairpinRagion])
					else:
						distanceToLoop2 = min([startPos2-item[1]-1 for item in hairpinRagion])
				else:
					for item in hairpinRagion:
						if len(set(range(startPos2, endPos2+1)) & set(range(item[0], item[1]+1))) == overlapLen2:
							if armType2 == 'arm5':
								distanceToLoop2 = item[0]-1-endPos2
							else:
								distanceToLoop2 = startPos2 - item[1]-1
							break
				armTypeOverlenDic.update({armType2: [overlapLen2, distanceToLoop2]})
			# Calculate the stem length of precusor.
			stemLen = 0
			for d in elementDic.keys():
				if d[0] == 's':
					stemLen = stemLen + max([item[1]+1-item[0] for item in elementDic[d]])
			# Check whether apical UGU/UGUG motif exist in hairpin region
			flag = 'No'
			for d in elementDic.keys():
				if d[0] == 'h':
					for item in elementDic[d]:
						seqTmp = self.premiRNASeq[item[0]:item[1]]
						if 'UGU' in seqTmp or 'UGUG' in seqTmp:
							flag = 'Yes'
							break
			# pair state
			if len(self.miRNASeq) == 2:
				pairState = 'Yes'
			else:
				pairState = 'No'
		else:
			hairpinCount = None
			bindingCount = None
			interiorLoopCount = None
			armType = None
			apicalLoopSize = None
			armTypeOverlenDic = None
			stemLen= None
			flag= None
			pairState = None
		return (hairpinCount, bindingCount, interiorLoopCount, armType, apicalLoopSize, armTypeOverlenDic, stemLen, flag, self.mfe, pairState)

	def getPrunedPrecusor_old(self):
		if len(self.miRNASeq) == 1:
			miRNASeq1 = self.miRNASeq[0]
			selectedMiRNASeq = miRNASeq1
		else:
			miRNASeq1 = self.miRNASeq[0]
			miRNASeq2 = self.miRNASeq[1]
			startPos1, endPos1 = getMiRNAPosition(miRNASeq1, self.premiRNASeq)
			if startPos1 == 20:
				selectedMiRNASeq = miRNASeq1
			else:
				selectedMiRNASeq = miRNASeq2
		armType = checkArm(self.premiRNASeq, self.premiRNAStructure, selectedMiRNASeq)
		bg1 = fgb.BulgeGraph()
		fa = '\n'.join(['>PrecusorStr', self.premiRNASeq, self.premiRNAStructure])
		bg1.from_fasta(fa)
		if armType == 'arm5':
			startPos1, endPos1 = getMiRNAPosition(selectedMiRNASeq, self.premiRNASeq)
			startPos1 = 0
			otherCount = 0 
			for i in range(startPos1, endPos1):
				if self.premiRNAStructure[i] == '(':
					break
				otherCount = otherCount + 1
			# Locate the partner of a base pair
			try:
				pairedPos = bg1.pairing_partner(startPos1+otherCount+1) - 1
				if otherCount + pairedPos >= len(self.premiRNASeq):
					newEndPos = len(self.premiRNASeq) - 1
				else:
					newEndPos = otherCount + pairedPos
				prunedPrecusorSeq = self.premiRNASeq[startPos1 : newEndPos+1]
				prunedPrecusorStr = self.premiRNAStructure[startPos1 : newEndPos+1]
			except TypeError:
				prunedPrecusorSeq = None
				prunedPrecusorStr = None
		else:
			startPos1, endPos1 = getMiRNAPosition(selectedMiRNASeq, self.premiRNASeq)
			endPos1 = len(self.premiRNASeq)
			print(startPos1)
			print(endPos1)
			otherCount = 0 
			for i in range(endPos1-1, -1, -1):
				if self.premiRNAStructure[i] == ')':
					break
				otherCount = otherCount + 1
			# Locate the partner of a base pair
			print(otherCount)
			try:
				pairedPos = bg1.pairing_partner(endPos1-otherCount) - 1
				if pairedPos - otherCount < 0:
					newStartPos = 0
				else:
					newStartPos = pairedPos - otherCount
				prunedPrecusorSeq = self.premiRNASeq[newStartPos : endPos1]
				prunedPrecusorStr = self.premiRNAStructure[newStartPos : endPos1]
			except TypeError:
				prunedPrecusorSeq = None
				prunedPrecusorStr = None
		# Decorate the CorePrecusor structure.
		if not (prunedPrecusorStr is None):
			prunedPrecusorStr = decorate_str(prunedPrecusorStr)
		return (prunedPrecusorSeq, prunedPrecusorStr)

	def featuresInPrunedPrecusor(self):
		prunedPrecusorSeq, prunedPrecusorStr = self.getPrunedPrecusor()
		if not (prunedPrecusorStr is None) and '(' in prunedPrecusorStr and ')' in prunedPrecusorStr:
			bg = fgb.BulgeGraph()
			fa = '\n'.join(['>prunedPrecusorStr', prunedPrecusorStr, prunedPrecusorStr])
			bg.from_fasta(fa)
			bindingCount1 = 0
			bindingCount2 = 0
			for item in prunedPrecusorStr:
				if item == '(' :
					bindingCount1 = bindingCount1 + 1
				if item == ')' :
					bindingCount2 = bindingCount2 + 1
			# Calculate bindingCount in the CorePrecusor
			bindingCount = max([bindingCount1, bindingCount2])
			# Calculate the hairpinCount and interiorLoopCount
			hairpinCount = 0
			interiorLoopCount = 0
			for string in bg.to_bg_string().strip().split('\n'):
				if 'define' in string:
					contentTmp = string.split(' ')
					elementType = contentTmp[1]
					if elementType[0] == 'h':
						hairpinCount = hairpinCount + 1
					if elementType[0] == 'i':
						interiorLoopCount = interiorLoopCount + 1
			# Calculate apicalLoopSize of the CorePrecusor
			# To get compatible with the older version of forgi, the below statement is abandoned.
			#elementTypeSeq, indeSeq = bg.to_element_string(with_numbers=True).split('\n')
			elementTypeSeq, indeSeq = to_element_string(bg, with_numbers=True).split('\n')
			
			elementIndexList = [elementTypeSeq[i]+indeSeq[i] for i in range(len(elementTypeSeq))]
			elementSeqList = []
			elementlocationList = []
			# the position is 0-based
			for index, item in enumerate(elementIndexList):
				if len(elementlocationList) == 0:
					elementSeqList.append(item)
					elementlocationList.append([index, index])
				elif item != elementSeqList[-1]:
					elementSeqList.append(item)
					elementlocationList.append([index, index])
				else:
					elementlocationList[-1][1] = index
			elementDic = {}
			for i in range(len(elementSeqList)):
				if elementSeqList[i] not in elementDic.keys():
					elementDic.update({elementSeqList[i] : [elementlocationList[i]]})
				else:
					elementDic[elementSeqList[i]].append(elementlocationList[i])
			hairpinRagion = []
			for d in elementDic.keys():
				if d[0] == 'h':
					hairpinRagion = hairpinRagion + elementDic[d]
			if len(hairpinRagion) >= 1:
				apicalLoopSize = max([len(range(item[0], item[1]+1)) for item in hairpinRagion])
			else:
				apicalLoopSize = 0
			armTypeOverlenDic = {}
			# Calculate the armType of the miRNA
			armType = checkArm(self.premiRNASeq, self.premiRNAStructure, self.miRNASeq[0])
			# Calculate the overlap of miRNA with the hairpin
			# startPos and endPos are the locations of the miRNA in the corePrecusor. 0-based
			startPos, endPos = getMiRNAPosition(self.miRNASeq[0], prunedPrecusorSeq)
			endPos = endPos - 1
			if len(hairpinRagion) >= 1:
				overlapLen = max([len(set(range(startPos, endPos+1)) & set(range(item[0], item[1]+1))) for item in hairpinRagion])
			else:
				overlapLen = 0
			if overlapLen == 0:
				if armType == 'arm5':
					distanceToLoop = min([item[0]-1-endPos for item in hairpinRagion])
				else:
					distanceToLoop = min([startPos-item[1]-1 for item in hairpinRagion])
			else:
				for item in hairpinRagion:
					if len(set(range(startPos, endPos+1)) & set(range(item[0], item[1]+1))) == overlapLen:
						if armType == 'arm5':
							distanceToLoop = item[0]-1-endPos
						else:
							distanceToLoop = startPos - item[1] - 1
						break
			armTypeOverlenDic.update({armType: [overlapLen, distanceToLoop]})
			if len(self.miRNASeq) == 2:
				armType2 = checkArm(self.premiRNASeq, self.premiRNAStructure, self.miRNASeq[1])
				startPos2, endPos2 = getMiRNAPosition(self.miRNASeq[1], prunedPrecusorSeq)
				endPos2 = endPos2 - 1
				if len(hairpinRagion) >= 1:
					overlapLen2 = max([len(set(range(startPos2, endPos2+1)) & set(range(item[0], item[1]+1))) for item in hairpinRagion])
				else:
					overlapLen2 = 0
				if overlapLen2 == 0:
					if armType2 == 'arm5':
						distanceToLoop2 = min([item[0]-1-endPos2 for item in hairpinRagion])
					else:
						distanceToLoop2 = min([startPos2-item[1]-1 for item in hairpinRagion])
				else:
					for item in hairpinRagion:
						if len(set(range(startPos2, endPos2+1)) & set(range(item[0], item[1]+1))) == overlapLen2:
							if armType2 == 'arm5':
								distanceToLoop2 = item[0]-1-endPos2
							else:
								distanceToLoop2 = startPos2 - item[1]-1
							break
				armTypeOverlenDic.update({armType2: [overlapLen2, distanceToLoop2]})
			# Calculate the stem length of precusor.
			stemLen = 0
			for d in elementDic.keys():
				if d[0] == 's':
					stemLen = stemLen + max([item[1]+1-item[0] for item in elementDic[d]])
			# Check whether apical UGU/UGUG motif exist in hairpin region
			flag = 'No'
			for d in elementDic.keys():
				if d[0] == 'h':
					for item in elementDic[d]:
						seqTmp = prunedPrecusorSeq[item[0]:item[1]]
						if 'UGU' in seqTmp or 'UGUG' in seqTmp:
							flag = 'Yes'
							break
			# minimun free energy
			with open(os.path.join(self.tmpdir, 'prunedPrecusor.fa'), 'w') as outf:
				fa_tmp = '\n'.join(['>'+self.premiRNASeqName, prunedPrecusorSeq, prunedPrecusorStr])
				outf.write(fa_tmp+'\n')
			f1 = os.path.join(self.tmpdir, 'prunedPrecusor.fa')
			f2 = os.path.join(self.tmpdir, 'prunedPrecusor.str')
			os.system('%s < %s --noPS --noLP > %s'%(self.rnafoldCmdTmp, f1, f2))
			with open(f2, 'r') as inf:
				try:
					tmp = inf.readlines()
					if len(tmp[-1].strip().split(' ')[1]) >= 3:
						mFE = float(tmp[-1].strip().split(' ')[1][1:-1].strip())
					else:
						mFE = float(tmp[-1].strip().split(' ')[2][:-1].strip())
				except:
					mFE = 0
			os.remove(f1)
			os.remove(f2)
			# pair state
			if len(self.miRNASeq) == 2:
				pairState = 'Yes'
			else:
				pairState = 'No'
		else:
			hairpinCount = None
			bindingCount = None
			interiorLoopCount = None
			armType = None
			apicalLoopSize = None
			armTypeOverlenDic = None
			stemLen= None
			flag= None
			mFE = None
			pairState = None
		return (hairpinCount, bindingCount, interiorLoopCount, armType, apicalLoopSize, armTypeOverlenDic, stemLen, flag, mFE, pairState)

	def percentageOfPairedInMiRNA(self):
		miRNAStr = getMiRNAStructure(self.miRNASeq[0], self.premiRNASeq, self.premiRNAStructure)
		try:
			percentage = (miRNAStr.count('(')+miRNAStr.count(')'))/float(len(miRNAStr))
		except:
			percentage = 0
		return percentage

	def bindingsInMiRNA(self):
		miRNAStr = getMiRNAStructure(self.miRNASeq[0], self.premiRNASeq, self.premiRNAStructure)
		countBindingsInMiRNA = miRNAStr.count('(')+miRNAStr.count(')')
		return countBindingsInMiRNA

	def getCorePrecusor(self):
		if len(self.miRNASeq) == 1:
			miRNASeq1 = self.miRNASeq[0]
			armType = checkArm(self.premiRNASeq, self.premiRNAStructure, miRNASeq1)
			bg1 = fgb.BulgeGraph()
			fa = '\n'.join(['>PrecusorStr', self.premiRNASeq, self.premiRNAStructure])
			bg1.from_fasta(fa)
			if armType == 'arm5':
				startPos1, endPos1 = getMiRNAPosition(miRNASeq1, self.premiRNASeq)
				otherCount = 0 
				for i in range(startPos1, endPos1):
					if self.premiRNAStructure[i] == '(':
						break
					otherCount = otherCount + 1
				# Locate the partner of a base pair
				try:
					pairedPos = bg1.pairing_partner(startPos1+otherCount+1) - 1
					if otherCount + pairedPos >= len(self.premiRNASeq):
						newEndPos = len(self.premiRNASeq) - 1
					else:
						newEndPos = otherCount + pairedPos
					corePrecusorSeq = self.premiRNASeq[startPos1 : newEndPos+1]
					corePrecusorStr = self.premiRNAStructure[startPos1 : newEndPos+1]
				except TypeError:
					corePrecusorSeq = None
					corePrecusorStr = None
			else:
				startPos1, endPos1 = getMiRNAPosition(miRNASeq1, self.premiRNASeq)
				otherCount = 0 
				for i in range(endPos1-1, -1, -1):
					if self.premiRNAStructure[i] == ')':
						break
					otherCount = otherCount + 1
				# Locate the partner of a base pair
				try:
					pairedPos = bg1.pairing_partner(endPos1-otherCount) - 1
					if pairedPos - otherCount < 0:
						newStartPos = 0
					else:
						newStartPos = pairedPos - otherCount
					corePrecusorSeq = self.premiRNASeq[newStartPos : endPos1]
					corePrecusorStr = self.premiRNAStructure[newStartPos : endPos1]
				except TypeError:
					corePrecusorSeq = None
					corePrecusorStr = None
		else:
			miRNASeq1 = self.miRNASeq[0]
			miRNASeq2 = self.miRNASeq[1]
			startPos1, endPos1 = getMiRNAPosition(miRNASeq1, self.premiRNASeq)
			startPos2, endPos2 = getMiRNAPosition(miRNASeq2, self.premiRNASeq)
			newStartPos = min([startPos1, endPos1, startPos2, endPos2])
			newEndPos = max([startPos1, endPos1, startPos2, endPos2])
			corePrecusorSeq = self.premiRNASeq[newStartPos : newEndPos]
			corePrecusorStr = self.premiRNAStructure[newStartPos : newEndPos]
		# Decorate the CorePrecusor structure.
		if not (corePrecusorStr is None):
			corePrecusorStr = decorate_str(corePrecusorStr)
		return (corePrecusorSeq, corePrecusorStr)

	def featuresInCorePrecusor(self):
		corePrecusorSeq, corePrecusorStr = self.getCorePrecusor()
		if not (corePrecusorStr is None) and '(' in corePrecusorStr and ')' in corePrecusorStr:
			bg = fgb.BulgeGraph()
			fa = '\n'.join(['>corePrecusorStr', corePrecusorSeq, corePrecusorStr])
			bg.from_fasta(fa)
			bindingCount1 = 0
			bindingCount2 = 0
			for item in corePrecusorStr:
				if item == '(' :
					bindingCount1 = bindingCount1 + 1
				if item == ')' :
					bindingCount2 = bindingCount2 + 1
			# Calculate bindingCount in the CorePrecusor
			bindingCount = max([bindingCount1, bindingCount2])
			# Calculate the hairpinCount and interiorLoopCount
			hairpinCount = 0
			interiorLoopCount = 0
			for string in bg.to_bg_string().strip().split('\n'):
				if 'define' in string:
					contentTmp = string.split(' ')
					elementType = contentTmp[1]
					if elementType[0] == 'h':
						hairpinCount = hairpinCount + 1
					if elementType[0] == 'i':
						interiorLoopCount = interiorLoopCount + 1
			# Calculate apicalLoopSize of the CorePrecusor
			# To get compatible with the older version of forgi, the below statement is abandoned.
			#elementTypeSeq, indeSeq = bg.to_element_string(with_numbers=True).split('\n')
			elementTypeSeq, indeSeq = to_element_string(bg, with_numbers=True).split('\n')
			
			elementIndexList = [elementTypeSeq[i]+indeSeq[i] for i in range(len(elementTypeSeq))]
			elementSeqList = []
			elementlocationList = []
			# the position is 0-based
			for index, item in enumerate(elementIndexList):
				if len(elementlocationList) == 0:
					elementSeqList.append(item)
					elementlocationList.append([index, index])
				elif item != elementSeqList[-1]:
					elementSeqList.append(item)
					elementlocationList.append([index, index])
				else:
					elementlocationList[-1][1] = index
			elementDic = {}
			for i in range(len(elementSeqList)):
				if elementSeqList[i] not in elementDic.keys():
					elementDic.update({elementSeqList[i] : [elementlocationList[i]]})
				else:
					elementDic[elementSeqList[i]].append(elementlocationList[i])
			hairpinRagion = []
			for d in elementDic.keys():
				if d[0] == 'h':
					hairpinRagion = hairpinRagion + elementDic[d]
			if len(hairpinRagion) >= 1:
				apicalLoopSize = max([len(range(item[0], item[1]+1)) for item in hairpinRagion])
			else:
				apicalLoopSize = 0
			armTypeOverlenDic = {}
			# Calculate the armType of the miRNA
			armType = checkArm(self.premiRNASeq, self.premiRNAStructure, self.miRNASeq[0])
			# Calculate the overlap of miRNA with the hairpin
			# startPos and endPos are the locations of the miRNA in the corePrecusor. 0-based
			startPos, endPos = getMiRNAPosition(self.miRNASeq[0], corePrecusorSeq)
			endPos = endPos - 1
			if len(hairpinRagion) >= 1:
				overlapLen = max([len(set(range(startPos, endPos+1)) & set(range(item[0], item[1]+1))) for item in hairpinRagion])
			else:
				overlapLen = 0
			if overlapLen == 0:
				if armType == 'arm5':
					distanceToLoop = min([item[0]-1-endPos for item in hairpinRagion])
				else:
					distanceToLoop = min([startPos-item[1]-1 for item in hairpinRagion])
			else:
				for item in hairpinRagion:
					if len(set(range(startPos, endPos+1)) & set(range(item[0], item[1]+1))) == overlapLen:
						if armType == 'arm5':
							distanceToLoop = item[0]-1-endPos
						else:
							distanceToLoop = startPos - item[1] - 1
						break
			armTypeOverlenDic.update({armType: [overlapLen, distanceToLoop]})
			if len(self.miRNASeq) == 2:
				armType2 = checkArm(self.premiRNASeq, self.premiRNAStructure, self.miRNASeq[1])
				startPos2, endPos2 = getMiRNAPosition(self.miRNASeq[1], corePrecusorSeq)
				endPos2 = endPos2 - 1
				if len(hairpinRagion) >= 1:
					overlapLen2 = max([len(set(range(startPos2, endPos2+1)) & set(range(item[0], item[1]+1))) for item in hairpinRagion])
				else:
					overlapLen2 = 0
				if overlapLen2 == 0:
					if armType2 == 'arm5':
						distanceToLoop2 = min([item[0]-1-endPos2 for item in hairpinRagion])
					else:
						distanceToLoop2 = min([startPos2-item[1]-1 for item in hairpinRagion])
				else:
					for item in hairpinRagion:
						if len(set(range(startPos2, endPos2+1)) & set(range(item[0], item[1]+1))) == overlapLen2:
							if armType2 == 'arm5':
								distanceToLoop2 = item[0]-1-endPos2
							else:
								distanceToLoop2 = startPos2 - item[1]-1
							break
				armTypeOverlenDic.update({armType2: [overlapLen2, distanceToLoop2]})
			# Calculate the stem length of precusor.
			stemLen = 0
			for d in elementDic.keys():
				if d[0] == 's':
					stemLen = stemLen + max([item[1]+1-item[0] for item in elementDic[d]])
			# Check whether apical UGU/UGUG motif exist in hairpin region
			flag = 'No'
			for d in elementDic.keys():
				if d[0] == 'h':
					for item in elementDic[d]:
						seqTmp = corePrecusorSeq[item[0]:item[1]]
						if 'UGU' in seqTmp or 'UGUG' in seqTmp:
							flag = 'Yes'
							break
			# minimun free energy
			with open(os.path.join(self.tmpdir, 'corePrecusor.fa'), 'w') as outf:
				fa_tmp = '\n'.join(['>'+self.premiRNASeqName, corePrecusorSeq, corePrecusorStr])
				outf.write(fa_tmp+'\n')
			f1 = os.path.join(self.tmpdir, 'corePrecusor.fa')
			f2 = os.path.join(self.tmpdir, 'corePrecusor.str')
			os.system('%s < %s --noPS > %s'%(self.rnafoldCmdTmp, f1, f2))
			with open(f2, 'r') as inf:
				try:
					tmp = inf.readlines()
					if len(tmp[-1].strip().split(' ')[1]) >= 3:
						mFE = float(tmp[-1].strip().split(' ')[1][1:-1].strip())
					else:
						mFE = float(tmp[-1].strip().split(' ')[2][:-1].strip())
				except:
					mFE = 0
			os.remove(f1)
			os.remove(f2)
			# pair state
			if len(self.miRNASeq) == 2:
				pairState = 'Yes'
			else:
				pairState = 'No'
		else:
			hairpinCount = None
			bindingCount = None
			interiorLoopCount = None
			armType = None
			apicalLoopSize = None
			armTypeOverlenDic = None
			stemLen= None
			flag= None
			mFE = None
			pairState = None
		return (hairpinCount, bindingCount, interiorLoopCount, armType, apicalLoopSize, armTypeOverlenDic, stemLen, flag, mFE, pairState)



#!/usr/bin/env python

#Built-in libraries 
from pathlib import Path
import time
import sys
import os
import multiprocessing
import pickle

# GitHub libraries
import pandas as pd

#Custom miRge libraries 
from mirge.libs.parse import parseArg
from mirge.libs.miRgeEssential import check_dependencies, validate_files
from mirge.libs.digest import baking 
from mirge.libs.digestEC import bakingEC
from mirge.libs.summary import summarize
from mirge.libs.manifoldAlign import bwtAlign
from mirge.libs.novel_mir import predict_nmir
from mirge.classes.exportHTML import FormatHTML

def main():
    globalstart = time.perf_counter()     
    args = parseArg()
    samples = args.samples
    if args.outDirName:
        ourDir_n = str(args.outDirName)
        workDir = Path(args.outDir)/ourDir_n if args.outDir else Path.cwd()/ourDir_n
    else:
        tStamp = time.strftime('%Y-%m-%d_%H-%M-%S',time.localtime(time.time()))
        ourDir = "miRge." + tStamp
        workDir = Path(args.outDir)/ourDir if args.outDir else Path.cwd()/ourDir
    Path(workDir).mkdir(exist_ok=True, parents=True)
    db_keys = {"mirbase":"miRBase", "mirgenedb":"MirGeneDB"}
    runlogFile = Path(workDir)/"run.log"
    outlog = open(str(runlogFile),"a+")
    outlog.write(" ".join(sys.argv))
    outlog.write("\n")
    outlog.close()
    check_dependencies(args, str(runlogFile))
    outlog = open(str(runlogFile),"a+")
    if args.tRNA_frag and args.organism_name != "human":
        outlog.write("ERROR: Detection of tRF(tRNA fragments) is only supported for human.\n")
        sys.exit("ERROR: Detection of tRF(tRNA fragments) is only supported for human.")

    indexFiles = Path(args.libraries_path)/args.organism_name/"index.Libs"
    isExist = os.path.exists(indexFiles)
    if not isExist:
        print("\n ERROR: The path to miRge libraries is incorrect or does not exist!\n")
        outlog = open(str(runlogFile),"a+")
        outlog.write("\n ERROR: The path to miRge libraries is incorrect or does not exist!\n")
        outlog.close()
        exit()

    if args.threads == 0:
        args.threads = multiprocessing.cpu_count()

    ref_db = db_keys.get(args.mir_DB.lower()) if args.mir_DB.lower() in db_keys else sys.exit("ERROR: Require valid database (-d miRBase or MirGeneDB)")
    if args.organism_name == "hamster":
        if "mirbase" in args.mir_DB.lower():
            print("Library for hamster is not developed for miRBase, therefore, MirGeneDB is used\n")
        ref_db = "MirGeneDB"
    if len(args.adapters) == 2:
        back = list(args.adapters[0])
        if back[1] == "illumina":
            back[1] = 'TGGAATTCTCGGGTGCCAAGGAACTCCAG'
        args.adapters[0] = tuple(back)

        front = list(args.adapters[1])
        if front[1] == "illumina":
            front[1] = 'GTTCAGAGTTCTACAGTCCGACGATC'
        args.adapters[1] = tuple(front)

    if len(args.adapters) == 1:
        somewhere = list(args.adapters[0])
        if somewhere[0] == "back" and somewhere[1] == "illumina":
            somewhere[1] = 'TGGAATTCTCGGGTGCCAAGGAACTCCAG'
            args.adapters[0] = tuple(somewhere)
        elif somewhere[0] == "front" and somewhere[1] == "illumina": 
            somewhere[1] = 'GTTCAGAGTTCTACAGTCCGACGATC'
            args.adapters[0] = tuple(somewhere)
    #print(args.adapters)
    if not args.quiet:
        print("Collecting and validating input files...")
    outlog.write("Collecting and validating input files...\n")
    file_exts = ['.txt', '.csv']
    file_list = samples[0].split(',')
    if Path(file_list[0]).is_dir():
        if args.resume:
            rootToPKL = str(Path(file_list[0]).absolute())
            pdf = str(Path(rootToPKL)/'collapsed.pkl')
            pdfacc = str(Path(rootToPKL)/'collapsed_accessories.pkl')
            if not Path(pdf).exists() and not Path(pdfacc).exists():
                print("\nERROR: The provided path doesn't contain pickle files with .pkl extensions! Please refer to documentation.\n")
                outlog.write("\nERROR: The provided path doesn't contain pickle files with .pkl extensions! Please refer to documentation.\n")
                outlog.close()
                exit()
            else:
                pdDataFrame = pd.read_pickle(pdf)
                with open(pdfacc, 'rb') as pklin:
                    bpkl = pickle.load(pklin)
                sampleReadCounts = bpkl[0]
                trimmedReadCounts = bpkl[1]
                trimmedReadCountsUnique = bpkl[2]
                fastq_fullPath = bpkl[3]
                base_names = bpkl[4]
        else:
            file_list = [str(x) for x in Path(file_list[0]).iterdir() if x.is_file()]
            file_list = sorted(file_list)
            fastq_fullPath,base_names = validate_files(args, file_list, str(runlogFile))
    elif Path(file_list[0]).exists() and Path(file_list[0]).suffix in file_exts: # READ TXT OR CSV FILE HERE
        with open(file_list[0]) as file:
            lines = [line.strip() for line in file]
            fastq_fullPath, base_names = validate_files(args, lines, str(runlogFile))
    else:  # READ FASTQ OR FASTQ.gz FILES HERE
        fastq_fullPath, base_names = validate_files(args, file_list, str(runlogFile))
    if not args.quiet and not args.resume:
        print(f"\nmiRge3.0 will process {len(fastq_fullPath)} out of {len(file_list)} input file(s).\n")
        outlog.write(f"\nmiRge3.0 will process {len(fastq_fullPath)} out of {len(file_list)} input file(s).\n\n")
        outlog.close()
    elif args.resume:
        print(f"\nmiRge3.0 will process {len(fastq_fullPath)} saved run(s) from binary pickle file.\n")
        outlog.write(f"\nmiRge3.0 will process {len(fastq_fullPath)} saved run(s) from binary pickle file.\n\n")
        outlog.close()


    if args.miREC and not args.resume:
        if args.uniq_mol_ids:
        #if args.umi:
            outlog = open(str(runlogFile),"a+")
            print(f"\nCurrently, error correction does not apply for the data with UMI sequences\n")
            outlog.write(f"\nCurrently, error correction does not apply for the data with UMI sequences\n")
            outlog.close()
            exit()
        else:
            pdDataFrame,sampleReadCounts,trimmedReadCounts,trimmedReadCountsUnique = bakingEC(args, fastq_fullPath, base_names, workDir)
    elif not args.resume:
        pdDataFrame,sampleReadCounts,trimmedReadCounts,trimmedReadCountsUnique = baking(args, fastq_fullPath, base_names, workDir)

    if args.save_pkl:
        pickleFile = str(Path(workDir)/'collapsed.pkl')
        pickleFile_acc = str(Path(workDir)/'collapsed_accessories.pkl')
        pdDataFrame.to_pickle(pickleFile)
        a = [sampleReadCounts, trimmedReadCounts,trimmedReadCountsUnique,fastq_fullPath, base_names]
        with open(pickleFile_acc, 'wb') as pklac:
            pickle.dump(a, pklac, protocol=pickle.HIGHEST_PROTOCOL)
    
    if args.save_pkl == args.resume and not args.save_pkl != "False":
        outlog = open(str(runlogFile),"a+")
        print(f"\nERROR: The arguments -pkl and -rr are mutually exclusive! Please refer to documentation.\n" )
        outlog.write(f"\nERROR: The arguments -pkl and -rr are mutually exclusive! Please refer to documentation.\n")
        outlog.close()
        exit()

    pdDataFrame = bwtAlign(args,pdDataFrame,workDir,ref_db)
    outlog = open(str(runlogFile),"a+")
    if not args.quiet:
        print(f"Summarizing and tabulating results...")
    outlog.write("\nSummarizing and tabulating results...\n")
    outlog.close()
    summary_Start_time = time.perf_counter()
    pdMapped = pdDataFrame[pdDataFrame.annotFlag.eq(1)]
    pdUnmapped = pdDataFrame[pdDataFrame.annotFlag.eq(0)]
    summarize(args, workDir, ref_db, base_names, pdMapped, sampleReadCounts, trimmedReadCounts, trimmedReadCountsUnique)

    #fileToCSV = Path(workDir)/"miRge3_collapsed.csv"
    mappedfileToCSV = Path(workDir)/"mapped.csv"
    unmappedfileToCSV = Path(workDir)/"unmapped.csv"
    #pdDataFrame.to_csv(fileToCSV)
    pdMapped.to_csv(mappedfileToCSV)
    pdUnmapped.to_csv(unmappedfileToCSV)
    summary_End_time = time.perf_counter()
    """
    Enabling Visualization HTML format
    """
    html = FormatHTML(workDir)
    html.beginHTML()
    html.histReadLen(len(base_names))
    if args.gff_out:
        html.isomirsTab(len(base_names), True)
    else:
        html.isomirsTab(len(base_names), False)
    
    html.exprTab(len(base_names))
    
    if args.uniq_mol_ids:
        html.umiTab(len(base_names), True)
    else:
        html.umiTab(len(base_names), False)

    outlog = open(str(runlogFile),"a+")
    if not args.quiet:
        print(f'Summary completed in {round(summary_End_time-summary_Start_time, 4)} second(s)\n')     
    outlog.write(f"Summary completed in {round(summary_End_time-summary_Start_time, 4)} second(s)\n")
    if args.novel_miRNA:
        html.novelTab(1)
        if not args.quiet:
            print("Predicting novel miRNAs\n")
        outlog.write("Predicting novel miRNAs\n")
        outlog.close()
        predict_nmir(args, workDir, ref_db, base_names, pdUnmapped)
        outlog = open(str(runlogFile),"a+")
    else:
        html.novelTab(0)
    #    novelTab
    for fname in os.listdir(str(Path(workDir))):
        if fname.endswith('.sam'):
            try:
                allSamFiles = Path(workDir)/"*.sam"
                os.system('rm -r %s'%(allSamFiles))
                break
            except OSError:    
                pass
    html.closeHTML()
    if args.diffex:
        if args.metadata:
            miRCountsFile = str(Path(workDir)/"miR.Counts.csv")
            diffOutFile = str(Path(workDir)/"miR.DiffExpn.txt")
            diffRData = str(Path(workDir)/"miR.DESeq2.RData")
            metaDataFile = str(Path(args.metadata).resolve())
            if not args.quiet:
                print("Performing differential expression...\n")
            outlog.write("Performing differential expression...\n")
            outlog.close()
            outlog = open(str(runlogFile),"a+")
            RscriptDirTmp = Path(__file__).resolve().parents[0]
            RscriptDir = Path(RscriptDirTmp)/('rScripts')/('difxp_DESeq.R')
            outvpdf = Path(workDir)/('mirge_diffex.volcano.pdf')
            outpcapdf = Path(workDir)/('mirge_diffex.pca.pdf')
            os.system('Rscript %s %s %s %s %s %s %s'%(RscriptDir, miRCountsFile, metaDataFile, outvpdf, outpcapdf, diffOutFile, diffRData))
        else: 
            print("Metadata file is required\n")
            # Rscript  /home/arun/repositories/Project_120919/mirge/rScripts/difxp_DESeq.R miR.Counts.csv DESmetadata.csv 

    globalend_time = time.perf_counter()
    if not args.quiet:
        resultsDir = str(Path(workDir).absolute())
        print(f'\nThe path to ourput directory: {resultsDir}')
        print(f'\nThe analysis completed in {round(globalend_time-globalstart, 4)} second(s)\n')     
    outlog.write(f"\nThe path to ourput directory: {resultsDir}")
    outlog.write(f"\nThe analysis completed in {round(globalend_time-globalstart, 4)} second(s)\n")
    outlog.close()



if __name__ == '__main__':
    main()
    



#!/usr/bin/python

#import timeit, sys

import mirge.forgi.threedee.utilities.cytvec as ftuc
import mirge.forgi.utilities.debug as fud
import numpy as np
import math as m
import numpy.linalg as nl
import numpy.testing as nt
import random as rand
#import scipy as sp

null_array = np.array([0., 0., 0.])

x_array = np.array([1., 0., 0.])
y_array = np.array([0., 1., 0.])
z_array = np.array([0., 0., 1.])

# identity matrix
identity_matrix = np.array([x_array, y_array, z_array])

standard_basis = np.array([[1., 0., 0.], [0., 1., 0.], [0., 0., 1.]])
tau = 2 * m.pi

def get_inter_distances(vecs):
    '''
    Calculate all of the distances between the points of vecs.
    '''
    distances = []
    for i in range(len(vecs)):
        for j in range(i+1, len(vecs)):
            distances += [vec_distance(vecs[i], vecs[j])]

    return distances

def get_random_vector(mult=1.):
    return np.array([mult * rand.uniform(-1, 1), mult * rand.uniform(-1, 1), mult * rand.uniform(-1, 1)])

def get_orthogonal_unit_vector(vec):
    '''
    Return a vector orthogonal to vec.
    '''
    vec2 = get_non_colinear_unit_vector(vec)
    vec3 = np.cross(vec, vec2)
    return normalize(vec3)

def get_random_vector_pair(angle=rand.uniform(0, m.pi)):
    vec1 = get_random_vector()
    vec2 = get_non_colinear_unit_vector(vec1)
    rot_vec = np.cross(vec1, vec2)
    rotmat = rotation_matrix(rot_vec, angle)
    vec2 = np.dot(rotmat, vec1)
    return (vec1, vec2)

def get_double_alignment_matrix(vp1, vp2):
    '''
    Align two sets of two vectors onto each other.

    @param vp1: A pair of two vectors.
    @param vp2: Another pair of two vectors.
    '''
    angle1 = vec_angle(vp1[0], vp1[1])
    angle2 = vec_angle(vp2[0], vp2[1])

    nt.assert_allclose(angle1, angle2, rtol=1e-7, atol=1e-7)

    # Align the first two segments
    mat1 = get_alignment_matrix(vp1[0], vp2[0])

    # See where the second segment of the second set ends up
    # after the first alignment
    new_vp2_1 = np.dot(mat1, vp2[1])

    comp1 = np.cross(vp1[1], vp1[0])
    #comp1 = np.cross(vp1[0], vp1[1])
    comp2 = np.cross(vp1[0], comp1) # should be along the plane of vp1[0] and vp1[1]

    basis1 = create_orthonormal_basis(normalize(vp1[0]), normalize(comp2))
    rej2 = change_basis(new_vp2_1, basis1, standard_basis)

    angle = m.atan2(rej2[2], rej2[1])

    mat2 = rotation_matrix(vp1[0], angle)

    #return np.dot(mat1, mat2)
    return np.dot(mat2, mat1)

def get_alignment_matrix(vec1, vec2):
    '''
    Return a rotation matrix that will align vec1 along vec2.

    @param vec1: The target vector.
    @param vec2: The vector to be aligned.
    '''

    comp = np.cross(vec1, vec2)
    angle = vec_angle(vec1, vec2)

    return rotation_matrix(comp, angle)


def get_non_colinear_unit_vector(vec): 
    '''
    Get a unit vector that does not lie on the line defined by vec.

    This is done by creating a vector along the least represented axis in vec.

    @param vec: The vector under consideration.
    @return: A vector along an axis.
    '''
    absvec = [abs(v) for v in vec]
    m = min(absvec)
    ind = absvec.index(m) 
    unit = [0., 0., 0.] 
    unit[ind] = 1. 

    return np.array(unit)

def create_orthonormal_basis1(vec1, vec2=None, vec3=None):
    '''
    Create an orthonormal basis using the provided vectors.

    If more than one is provided, it must be orthogonal to 
    the others.
    '''
    if vec2 == None:
        vec2 = get_non_colinear_unit_vector(vec1)
    #else:
    #    nt.assert_allclose(np.dot(vec2, vec1), 0., rtol=1e-7, atol=1e-7)


    vec1 = normalize(np.array(vec1))
    vec2 = normalize(np.array(vec2))

    if vec3 == None:
        vec3 = np.cross(vec1, vec2)

    vec3 = normalize(vec3)

    return np.array([vec1, vec2, vec3])


def create_orthonormal_basis(vec1, vec2=None, vec3=None):
    '''
    Create an orthonormal basis using the provided vectors.

    If more than one is provided, it must be orthogonal to 
    the others.
    '''
    if vec2 == None:
        vec2 = get_non_colinear_unit_vector(vec1)
        vec2 = np.cross(vec1, vec2)
    #else:
    #    nt.assert_allclose(np.dot(vec2, vec1), 0., rtol=1e-7, atol=1e-7)

    vec1 /= magnitude(vec1)
    vec2 /= magnitude(vec2)

    if vec3 == None:
        vec3 = np.cross(vec1, vec2)

    vec3 /= magnitude(vec3)

    return np.array([vec1, vec2, vec3])

def time_cob1():
    vec1 = get_random_vector()
    vec2 = get_random_vector()

    basis = create_orthonormal_basis1(vec1, vec2)

def time_cob2():
    vec1 = get_random_vector()
    vec2 = get_random_vector()

    basis = create_orthonormal_basis(vec1, vec2)

def time_cob():
    t1 = timeit.Timer("time_cob1()", "from forgi.utilities.vector import time_cob1")
    t2 = timeit.Timer("time_cob2()", "from forgi.utilities.vector import time_cob2")

    print(t1.repeat(number=10000))
    print(t2.repeat(number=10000))

def spherical_cartesian_to_polar(vec):
    '''
    Return a parameterization of a vector of 3 coordinates:

    x = r sin u cos v
    y = r sin u sin v
    z = r cos u

    0 <= u <= pi
    -pi <= v <= pi

    Where u is the polar angle and v is the azimuth angle.

    @param vec: A vector of 3 cartesian coordinates.
    @return: (r, u, v)
    '''
    r = magnitude(vec)
    u = m.acos(vec[2] / r)
    v = m.atan2(vec[1], vec[0])

    nt.assert_allclose(vec[0], r * m.sin(u) * m.cos(v), rtol=1e-7, atol=1e-7)
    return (r, u, v)

def spherical_polar_to_cartesian(vec):
    '''
    Convert spherical polar coordinates to cartesian coordinates:

    See the definition of spherical_cartesian_to_polar.

    @param vec: A vector of the 3 polar coordinates (r, u, v)
    @return: (x, y, z)
    '''
    (r, u, v) = vec

    x = r * m.sin(u) * m.cos(v)
    y = r * m.sin(u) * m.sin(v)
    z = r * m.cos(u)

    return [x, y, z]

def get_standard_basis(dim):
    '''
    Get a standard basis for the given dimension.

    For 2D, this equals [[1.,0.],[0.,1.]]

    @param dim: The dimension of the vector space.
    @return: A vector of vectors that constitute the standard basis.
    '''

    standard_basis = [[0. for j in range(dim)] for i in range(dim)]
    for i in range(dim):
        standard_basis[i][i] = 1.
    standard_basis = np.array(standard_basis)

    return standard_basis

def change_basis(coords, new_basis, old_basis):
    '''
    Change the basis of coordinates to a new basis. In a regular structure
    we have the coordinates in the regular cartesian coordinate system. For helix-helix
    orientations, however, we want to express the coordinates in a coordinate system
    defined by the first helix.

    The new basis will consist of the axis of the first helix, one of its twist elements
    (the one closest to the second vector) and a third vector orthogonal to the previous
    two.

    # http://tutorial.math.lamar.edu/Classes/LinAlg/ChangeOfBasis.aspx

    @param coords: The coordinates to transform (array of n elements).
    @param new_basis: The new basis vectors (n x n matrix)
    @param old_basis: The old basis for the coordinates(n x n matrix)
    @return: The new coordinates according to the new basis
    '''
    #assert(len(coords) == len(new_basis))
    #assert(len(new_basis) == len(old_basis))

    dim = len(coords)
    #print "coords:", coords
    standard_coords = np.dot(old_basis.transpose(), coords)
    '''
    #print "standard_coords:", standard_coords
    standard_to_new = inv(new_basis.transpose())
    #print "standard_to_new:", standard_to_new
    new_coords = np.dot(standard_to_new, standard_coords)
    print "new_coords:", new_coords
    '''

    new_coords = nl.solve(new_basis.transpose(), standard_coords)
    #print "new_coords1:", new_coords1

    return new_coords


def change_basis1(coords, new_basis, old_basis):
    '''
    '''

    dim = len(coords)
    standard_coords = np.dot(old_basis.transpose(), coords)
    standard_to_new = nl.inv(new_basis.transpose())
    new_coords = np.dot(standard_to_new, standard_coords)

    return new_coords


def change_basis2(coords, new_basis, old_basis):
    '''
    '''

    dim = len(coords)
    standard_coords = np.dot(old_basis.T, coords)
    new_coords = nl.solve(new_basis.T, standard_coords)

    return new_coords

def change_basis1_benchmark():
    coords = get_random_vector(10.)
    basis1 = np.array([get_random_vector(10.) for i in range(3)])
    basis2 = np.array([get_random_vector(10.) for i in range(3)])

    nc = change_basis1(coords, basis1, basis2)

def change_basis2_benchmark():
    coords = get_random_vector(10.)
    basis1 = np.array([get_random_vector(10.) for i in range(3)])
    basis2 = np.array([get_random_vector(10.) for i in range(3)])

    nc = change_basis2(coords, basis1, basis2)

def time_basis1():
    t1 = timeit.Timer("change_basis1_benchmark()","from forgi.utilities.vector import change_basis1_benchmark")
    print(t1.repeat(number=10000))

def time_basis2():
    t2 = timeit.Timer("change_basis2_benchmark()","from forgi.utilities.vector import change_basis2_benchmark")
    print(t2.repeat(number=10000))

def vector_rejection(a, b):
    '''
    Return the vector rejection of a from b. In other words, return the orthogonal
    projection of a onto the plane orthogonal to b.

    @param a: The vector to be projected.
    @param b: The vector defining the normal of the plane.
    @return: The rejection of the vector a from b. (a - (np.dot(a, b) / np.dot(b, b)) * b)
    '''

    n = np.dot(a, b)
    d = np.dot(b, b)
    return a - (n / d) * b

def rotation_matrix_weave(axis, theta, mat = None):
    '''
    Calculate the rotation matrix for a rotation of theta degrees around axis.
    
    Implemented in C++ using the weave module. Runs approximately 6x faster than 
    the numpy version below if no mat is passed in and around 20x faster if mat is
    passed in.

    Thanks to unutbu on StackOverflow 

    http://stackoverflow.com/questions/6802577/python-rotation-of-3d-vector

    @param axis: The axis around which to rotate
    @param theta: The angle of rotation
    @return: A matrix which can be used to perform the given rotation. The coordinates
             need only be multiplied by the matrix.
    '''
    if mat == None:
        mat = np.eye(3,3)

    support = "#include <math.h>"
    code = """
        double x = sqrt(axis[0] * axis[0] + axis[1] * axis[1] + axis[2] * axis[2]);
        double a = cos(theta / 2.0);
        double b = -(axis[0] / x) * sin(theta / 2.0);
        double c = -(axis[1] / x) * sin(theta / 2.0);
        double d = -(axis[2] / x) * sin(theta / 2.0);

        mat[0] = a*a + b*b - c*c - d*d;
        mat[1] = 2 * (b*c - a*d);
        mat[2] = 2 * (b*d + a*c);

        mat[3*1 + 0] = 2*(b*c+a*d);
        mat[3*1 + 1] = a*a+c*c-b*b-d*d;
        mat[3*1 + 2] = 2*(c*d-a*b);

        mat[3*2 + 0] = 2*(b*d-a*c);
        mat[3*2 + 1] = 2*(c*d+a*b);
        mat[3*2 + 2] = a*a+d*d-b*b-c*c;
    """

    import scipy as sp
    sp.weave.inline(code, ['axis', 'theta', 'mat'], support_code = support, libraries = ['m'])

    return mat


def vector_set_rmsd(set1, set2):
    '''
    Calculate the rmsd between two sets of vectors.

    @param set1: A matrix
    @param set2: Another matrix.

    @return: The rmsd between the rows of the matrix.
    '''
    rmsd = 0
    count = 0
    for i in range(len(set1)):
        rmsd += magnitude(set2[i] - set1[i]) ** 2
        count += 1
    rmsd /= count
    return m.sqrt(rmsd)

def rotation_matrix(axis, theta):
    '''
    Calculate the rotation matrix for a rotation of theta degrees around axis.

    Thanks to unutbu on StackOverflow 

    http://stackoverflow.com/questions/6802577/python-rotation-of-3d-vector

    @param axis: The axis around which to rotate
    @param theta: The angle of rotation
    @return: A matrix which can be used to perform the given rotation. The coordinates
             need only be multiplied by the matrix.
    '''
    axis = axis/m.sqrt(np.dot(axis, axis))
    a = m.cos(theta/2)
    b, c, d = -axis*m.sin(theta/2)

    return np.array([[a*a+b*b-c*c-d*d, 2*(b*c-a*d), 2*(b*d+a*c)],
                  [2*(b*c+a*d), a*a+c*c-b*b-d*d, 2*(c*d-a*b)],
                  [2*(b*d-a*c), 2*(c*d+a*b), a*a+d*d-b*b-c*c]])

def get_vector_centroid(crds1):
    '''
    Find the centroid of a set of vectors.

    @param crds: A matrix containing all of the vectors.

    @return: The centroid of the rows of the matrix crds.
    '''
    centroid1 = np.array([0., 0., 0.])

    for i in range(len(crds1)):
        centroid1 += crds1[i]

    centroid1 /= float(len(crds1))

    for i in centroid1:
        if m.isnan(i):
            raise Exception('nan encountered')

    return centroid1

def center_on_centroid(crds1):
    centroid1 = get_vector_centroid(crds1)
    crds = []

    for i in range(len(crds1)):
        crds += [crds1[i] - centroid1]

    return np.array(crds)

def magnitude(vec):
    '''
    Return the magnitude of a vector (|V|).

    @param vec: The vector in question.
    @return: The magnitude of the vector.
    '''
    return ftuc.magnitude(vec)
    #return m.sqrt(np.dot(vec, vec))

def time_mag1():
    vec1 = get_random_vector()

    return m.sqrt(np.dot(vec1, vec1))

def time_mag2():
    vec1 = get_random_vector()

    return m.sqrt(np.dot(vec1, vec1))

def time_mag():
    t1 = timeit.Timer("time_mag1()", "from forgi.utilities.vector import time_mag1")
    t2 = timeit.Timer("time_mag2()", "from forgi.utilities.vector import time_mag2")

    print(t1.repeat(number=10000))
    print(t2.repeat(number=10000))


def normalize(vec):
    '''
    Normalize a vector so that its magnitude becomes 1.0 while
    its direction remains the same.

    @param vec: The vector in question.
    @return: A normalized version of the vector.
    '''

    return vec / magnitude(vec)

def vec_angle(vec1, vec2):
    '''
    Get the angle between two vectors using the identity:

    A * B = |A||B| cos t

    Where A and B are two vectors and t is the angle between them.

    @param vec1: The first vector (A)
    @param vec2: The second vector (B)
    @return: The angle between the two vectors.
    '''

    vec1n = normalize(vec1)
    vec2n = normalize(vec2)

    d = np.dot(vec1n, vec2n)

    # this shouldn't happen, but sometimes it does, presumably because
    # of rounding errors
    if d >= 1.:
        d = 1.
    if d <= -1.:
        d = -1.

    angle = m.acos(d)
    return angle

def vec_dot(a, b):
    return a[0] * b[0] + a[1] * b[1] + a[2] * b[2]

'''
def cross(a, b):
    c = [a[1]*b[2] - a[2]*b[1],
         a[2]*b[0] - a[0]*b[2],
         a[0]*b[1] - a[1]*b[0]]

    return c
'''

def vec_distance(vec1, vec2):
    return ftuc.vec_distance(vec1, vec2)
    #return m.sqrt(np.dot(vec2 - vec1, vec2 - vec1))

def line_segment_distance(s1_p0, s1_p1, s2_p0, s2_p1):
    '''
    Calculate the two points on each of the segments that are closest to
    each other. The first segment is defined as p1->p2 and the second as
    p3->p4.

    Code shamelessly translated from:
    http://softsurfer.com/Archive/algorithm_0106/algorithm_0106.htm#dist3D_Segment_to_Segment

    @param s1_p0: The start of the first segment
    @param s1_p1: The end of the first segment

    @param s2_p0: The start of the second segment
    @param s2_p1: The end of the second segment
    @return: A tuple of points (i1,i2) containing the point i1 on s1
        closest to the point i2 on segment s2
    '''
    u = s1_p1 - s1_p0
    v = s2_p1 - s2_p0
    w = s1_p0 - s2_p0
    
    a = np.dot(u,u)        # always >= 0
    b = np.dot(u,v)
    c = np.dot(v,v)        # always >= 0
    d = np.dot(u,w)
    e = np.dot(v,w)
    D = a*c - b*b       # always >= 0
    sD = D      # sc = sN / sD, default sD = D >= 0
    tD = D      # tc = tN / tD, default tD = D >= 0

    SMALL_NUM = 0.000001

    # compute the line parameters of the two closest points
    if (D < SMALL_NUM):  # the lines are almost parallel
        sN = 0.0        # force using point P0 on segment S1
        sD = 1.0        # to prevent possible division by 0.0 later
        tN = e
        tD = c
    else:                # get the closest points on the infinite lines
        sN = (b*e - c*d)
        tN = (a*e - b*d)
        if (sN < 0.0):      # sc < 0 => the s=0 edge is visible
            sN = 0.0
            tN = e
            tD = c
        elif (sN > sD):  # sc > 1 => the s=1 edge is visible
            sN = sD
            tN = e + b
            tD = c

    if (tN < 0.0):           # tc < 0 => the t=0 edge is visible
        tN = 0.0
        # recompute sc for this edge
        if (-d < 0.0):
            sN = 0.0
        elif (-d > a):
            sN = sD
        else:
            sN = -d
            sD = a
    elif (tN > tD):      # tc > 1 => the t=1 edge is visible
        tN = tD
        # recompute sc for this edge
        if ((-d + b) < 0.0):
            sN = 0
        elif ((-d + b) > a):
            sN = sD
        else:
            sN = (-d + b)
            sD = a
            
    # finally do the division to get sc and tc
    sc = 0.0 if abs(sN) < SMALL_NUM else sN / sD
    tc = 0.0 if abs(tN) < SMALL_NUM else tN / tD

    # get the difference of the two closest points
    dP = w + (sc * u) - (tc * v)  # = S1(sc) - S2(tc)

    return (s1_p0 + sc * u, s2_p0 + tc * v)

def closest_point_on_seg(seg_a, seg_b, circ_pos):
    '''
    Closest point between a line segment and a point.

    Lifted from:

    http://doswa.com/2009/07/13/circle-segment-intersectioncollision.html
    '''
    seg_v = seg_b - seg_a
    pt_v = circ_pos - seg_a
    mag = m.sqrt(sum(seg_v * seg_v))

    if m <= 0:
        raise ValueError("Invalid segment length")
    seg_v_unit = seg_v / mag
    proj = pt_v.dot(seg_v_unit)
    if proj <= 0:
        return seg_a.copy()
    if proj >= mag:
        return seg_b.copy()
    proj_v = seg_v_unit * proj
    closest = proj_v + seg_a
    return closest

def segment_circle(seg_a, seg_b, circ_pos, circ_rad):
    '''
    Something. Lifted from:

    http://doswa.com/2009/07/13/circle-segment-intersectioncollision.html
    '''
    closest = closest_point_on_seg(seg_a, seg_b, circ_pos)
    dist_v = circ_pos - closest
    mag = m.sqrt(sum(dist_v * dist_v))
    if m > circ_rad:
        return vec(0, 0)
    if mag(dist_v) <= 0:
        raise ValueError("Circle's center is exactly on segment")
    offset = dist_v / mag(dist_v) * (circ_rad - mag(dist_v))
    return offset

def cylinder_line_intersection(cyl, line, r):
    '''
    Get the points of intersection between a line and a cylinder.

    If they do not intersect, return an empty list. If the line
    touches the cylinder, then return a 2 point list with two identical points. 
    If the line crosses the cylinder, then return a list of 2 points.
    '''

    cyl = np.array(cyl)
    line = np.array(line)
    cyl_vec = cyl[1] - cyl[0]
    line_vec = line[1] - line[0]

    cyl_basis = create_orthonormal_basis(cyl_vec)

    line_t = change_basis(line.T, cyl_basis, standard_basis).T
    cyl_t = change_basis(cyl.T, cyl_basis, standard_basis).T

    cyl_vec_t = cyl_t[1] - cyl_t[0]
    line_vec_t = line_t[1] - line_t[0]

    line_vec_t_normed = normalize(line_vec_t)

    # get the point on the line closest to the
    # center of the cylinder
    p = closest_point_on_seg(line_t[0][1:], line_t[1][1:], 
                             np.array(cyl_t[0][1:]))

    if line_vec_t[1] == 0:
        return []

    # Figure out the x position by determining how far along
    # the y-coordinate of the segment the closest point is
    x  = (line_t[0][0] + 
            (line_vec_t[0] *
            (p[0] - line_t[0][1]) / line_vec_t[1]))
    v = p - cyl_t[0][1:]
    o = m.sqrt(sum(v * v))
    p = [x, p[0], p[1]]

    if o > r:
        # no intersection
        return []

    t = m.sqrt(r ** 2 - o ** 2)

    i1 = p - t * line_vec_t_normed
    i2 = p + t * line_vec_t_normed

    endpoints_t = [i1, i2]

    endpoints_t = sorted(endpoints_t, key=lambda x: x[0])
    line_t = sorted(line_t, key=lambda x: x[0])
    cyl_t = sorted(cyl_t, key=lambda x: x[0])

    # the start point will be the highest of the top of the cylinder
    # the end of the line, and the point where the line intersects
    # the surface of the cylinder
    start_points = np.array([cyl_t[0][0], line_t[0][0], endpoints_t[0][0]])

    # the end point will be the lowest of the... " " "
    end_points = np.array([cyl_t[1][0], line_t[1][0], endpoints_t[1][0]])

    start_points = sorted(start_points)
    end_points = sorted(end_points)

    start_param = (start_points[-1] - p[0]) / line_vec_t_normed[0]
    end_param = (end_points[0] - p[0]) / line_vec_t_normed[0]

    endpoints_t[0] = p + start_param * line_vec_t_normed
    endpoints_t[1] = p + end_param * line_vec_t_normed

    '''
    endpoints_t[0] = [start_points[-1],
                      endpoints_t[
    endpoints_t[1][0] = end_points[0]
    real_start = start_points[-1]
    real_end = end_points[0]
    '''

    intersects_t = np.array(endpoints_t)

    if intersects_t[0][0] > intersects_t[1][0]:
        # degenerate case, the line is to the side of the cylinder
        return []

    return change_basis(intersects_t.T, standard_basis, cyl_basis).T

def pin_fits_two_cyl(cyl1, cyl2, cyl_width):
    '''
    If we create a cone that starts at one end of cylinder1, does it
    enclose cylinder2?

    This function approximates an answer by projecting a circle on 
    the plane normal to the axis of cylinder1.

    @param cyl1: The coordinates of cylinder1
    @param cyl2: The coordinates of cylinder2
    @param cyl_width: The widths of the two cylinders
    @return: True or False
    '''
    basis = create_orthonormal_basis(cyl1[1] - cyl1[0])
    cyl2 = np.array([cyl2[0] - cyl1[1], cyl2[1] - cyl1[1]])
    
    cyl2_t = change_basis(cyl2.T, basis, standard_basis).T
    cone_width = cyl_width
    cyl1_len = magnitude(cyl1[1] - cyl1[0])
    
    # the cone expands
    cone_width_cyl2_start = cyl_width * (cyl1_len + cyl2_t[0][0]) / cyl1_len
    cone_width_cyl2_end = cyl_width * (cyl1_len + cyl2_t[1][0]) / cyl1_len
    
    cyl2_start_offset = m.sqrt(cyl2_t[0][1] ** 2 + cyl2_t[0][2] ** 2)
    cyl2_end_offset = m.sqrt(cyl2_t[1][1] ** 2 + cyl2_t[1][2] ** 2)
    
    if cyl2_start_offset > cone_width_cyl2_start:
        return False
    if cyl2_end_offset > cone_width_cyl2_end:
        return False
    
    return True

def point_in_cylinder(pt1, pt2, r, testpt):
    '''
    Determine if testpt is within a cylinder of radius r.

    Translated from:

        http://www.flipcode.com/archives/Fast_Point-In-Cylinder_Test.shtml

    @param pt1: The start of the cylinder axis.
    @param pt2: The end of the cylinder axis.
    @param r: The radius of the cylinder.
    @param testpt: The point we are testing.
    @return: True if the point is within the cylinder, False otherwise.
    '''
    d = pt2 - pt1
    lengthsq = np.dot(d, d)
    radius_sq = r * r

    pd = testpt - pt1
    dot = np.dot(pd, d)

    if dot < 0. or dot > lengthsq:
        # beyond the edges of the cylinder
        return False
    else:
        dsq = np.dot(pd, pd) - (dot * dot) / lengthsq

        if (dsq > radius_sq):
            return False
        else:
            return True



import random
import itertools as it

import mirge.forgi.utilities.debug as cud

def grouped(iterable, n):
    '''
    Return a list of every n elements in iterable.

    http://stackoverflow.com/questions/5389507/iterating-over-every-two-elements-in-a-list

    s -> (s0,s1,s2,...sn-1), (sn,sn+1,sn+2,...s2n-1), (s2n,s2n+1,s2n+2,...s3n-1), ...
    '''
    return it.izip(*[iter(iterable)]*n)

def merge_intervals(intervals, diff = 0):
    '''
    Take a set of intervals, and combine them whenever the endpoints
    match.

    I.e. [(42,47), (55,60), (60,63), (1,9), (63,71)]

    Should yield

    [(1,9),(42,47), (55,71)]

    There should be no overlapping intervals.

    @param intervals: A set of tuples indicating intervals
    @return: A list of merged intervals
    '''
    intervals.sort()
    iter_intervals = iter(intervals)

    # get the first interval
    curr_interval = list(next(iter_intervals))

    merged_intervals = []

    for i in iter_intervals:
        if abs(i[0] - curr_interval[1]) <= diff:
            # the start of this interval is equal to the end of the
            # current merged interval, so we merge it
            curr_interval[1] = i[1]
        else:
            # start a new interval and add the current merged one
            # to the list of intervals to return
            merged_intervals += [curr_interval]
            curr_interval = list(i)

    merged_intervals += [curr_interval]
    return merged_intervals

def gen_random_sequence(l):
    '''
    Generate a random RNA sequence of length l.
    '''
    return "".join([random.choice(['A','C','G','U']) for i in range(l)])



from Bio import SeqIO

def split_fasta_from_sam(infTmp1, infTmp2, imperfect_FASTA):
    # The arguments of this function are as follows:
    # mapped_mirna_SRR944031_vs_representative_seq_tmp1.sam  mapped_mirna_SRR944031.fa
    filterReadNameList = []
    with open(infTmp1, 'r') as inf1:
        for line in inf1:
            if line[0] != '@':
                temp = line.strip().split('\t')
                if temp[1] in ['0', '16']:
                    readName = line.strip().split('\t')[0]
                    if readName not in filterReadNameList:
                        filterReadNameList.append(readName)
    outf = open(imperfect_FASTA,'w')
    for record in SeqIO.parse(infTmp2, "fasta"):
        if record.id not in filterReadNameList:
            outf.write('>'+record.id+'\n')
            outf.write(str(record.seq)+'\n')
    outf.close()

def combineSam(infTmp1, infTmp2, combined_Sam):
    # The parameters of this function are as follows:
    # mapped_mirna_SRR944031_vs_representative_seq_tmp1.sam  mapped_mirna_SRR944031_vs_representative_seq_tmp2.sam
    outf = open(combined_Sam, 'w')
    with open(infTmp1,'r') as inf1:
        for line in inf1:
            if line[0] == '@':
                outf.write(line)
            else:
                temp = line.strip().split('\t')
                if temp[1] in ['0', '16']:
                    outf.write(line)
                else:
                    pass
    with open(infTmp2, 'r') as inf2:
        for line in inf2:
            if line[0] != '@':
                outf.write(line)
            else:
                pass
    outf.close()

def decorateSam(infTmp1, infTmp2, outfile_modifiedSam, *arg):
    # The parameters of this function are as follows:
    # unmapped_mirna_SRR944031_vs_representative_seq.sam unmapped_mirna_SRR944031.fa mapped_mirna_SRR944031_vs_genome_sorted_clusters_overlap14_trimmed_orig.fa
    seqDic = {}
    for record in SeqIO.parse(infTmp2, "fasta"):
        seqDic.update({record.id:str(record.seq)})
    outf = open(outfile_modifiedSam,"w")
    if len(arg) == 1:
        clusterSeqDic = {}
        for record in SeqIO.parse(arg[0], "fasta"):
            clusterSeqDic.update({record.id:str(record.seq)})
        with open(infTmp1,"r") as inf:
            for line in inf:
                if line[0] == '@':
                    outf.write(line)
                else:
                    miRNAName = line.strip().split('\t')[0]
                    clusterRNAName = line.strip().split('\t')[2]
                    if clusterRNAName in clusterSeqDic.keys():
                        outf.write('\t'.join([miRNAName, miRNAName.split('_')[-1], seqDic[miRNAName], clusterSeqDic[clusterRNAName]])+'\t'+'\t'.join(line.strip().split('\t')[1:])+'\n')
                    else:
                        outf.write('\t'.join([miRNAName, miRNAName.split('_')[-1], seqDic[miRNAName], '*'])+'\t'+'\t'.join(line.strip().split('\t')[1:])+'\n')
    else:
        with open(infTmp1,"r") as inf:
            for line in inf:
                if line[0] == '@':
                    outf.write(line)
                else:
                    miRNAName = line.strip().split('\t')[0]
                    outf.write('\t'.join([miRNAName, miRNAName.split('_')[-1], seqDic[miRNAName]])+'\t'+'\t'.join(line.strip().split('\t')[1:])+'\n') 
    outf.close()

def parse_refine_sam(infTmp, SelectTSV, RevKeptTSV):
    # The argument of this function is as follows:
    # unmapped_mirna_SRR944031_vs_representative_seq_bowtie2_modified.sam
    outf1 = open(SelectTSV,"w")
    outf2 = open(RevKeptTSV, "w")
    with open(infTmp,"r") as inf:
        for line in inf:
            if line[0] != "@":
                content = line.strip().split("\t")
                readName= content[0]
                readCounts = int(content[1])
                readSeq = content[2]
                clusterdSeq = content[3]
                # If the read in unmapped, the clusterdSeq will be "*"
                samFlags = content[4]
                clusterdName = content[5]
                alignedSeq = content[12]
                if samFlags in ["0", "256"]:
                    outf1.write(line)
                    outf2.write(line)
                elif samFlags in ["16", "272"]:
                    outf2.write(line)
                else:
                    pass
    outf1.close()
    outf2.close()



import os
import sys
from pathlib import Path
from Bio.Seq import Seq
# from Bio.Alphabet import generic_dna
from mirge.classes.readPrecusor import ReadPrecusor

def chunks(arr, n):
    return [arr[i:i+n] for i in range(0, len(arr), n)]

def isset(v): 
    try: 
        type (eval(v)) 
    except: 
        return 0 
    else: 
        return 1

def selcteOptimalPrecusor(precusorFeatureList, precusorList, clusterSeqType):
    #To do, select the precusor based on mFE or precusorOrignialMFE. Right now, I choose precusorOrignialMFE
    if clusterSeqType == 'clusterSeq':
        bindingCountThreshold = 19
    elif clusterSeqType == 'stableClusterSeq':
        bindingCountThreshold = 19-1-3
    else:
        bindingCountThreshold = 19
    if len(precusorFeatureList) != len(precusorList):
        print('Errors happen at precusorFeatureList and precusorList because of different length')
        sys.exit(1)
    retainedIndex = []
    nullIndex = []
    for i in range(len(precusorFeatureList)):
        if not (None in precusorFeatureList[i]):
            armTypeOverlenDic, hairpinCount, bindingCount, percent, mFE, precusorOrignialMFE = precusorFeatureList[i]
            if hairpinCount > 0 and bindingCount >= bindingCountThreshold and percent >= 0.6:
                if 'arm5' in armTypeOverlenDic.keys() and 'arm3' in armTypeOverlenDic.keys():
                    if armTypeOverlenDic['arm5'][0] <= 5 and armTypeOverlenDic['arm3'][0] == 0:
                        retainedIndex.append(i)
                elif 'arm5' in armTypeOverlenDic.keys():
                    if armTypeOverlenDic['arm5'][0] <= 5:
                        retainedIndex.append(i)
                elif 'arm3' in armTypeOverlenDic.keys():
                    if armTypeOverlenDic['arm3'][0] == 0:
                        retainedIndex.append(i)
                else:
                    pass
        else:
            nullIndex.append(i)

    if len(retainedIndex) >= 1:
        #To do, select the precusor based on mFE or precusorOrignialMFE. Right now precusorOrignialMFE is chosen.
        tmpList = [[precusorFeatureList[j][5], j] for j in retainedIndex]
        tmpList.sort()
        outputIndex = tmpList[0][1]
        selected_precusor = precusorList[outputIndex]
    else:
        newrangeList = []
        for i in range(len(precusorFeatureList)):
            if i not in nullIndex:
                newrangeList.append(i)
        if len(newrangeList) >= 1:
            tmpList = [[precusorFeatureList[j][5], j] for j in newrangeList]
            tmpList.sort()
            outputIndex = tmpList[0][1]
            selected_precusor = precusorList[outputIndex]
        else:
            outputIndex = None
            selected_precusor = None
    return (outputIndex, selected_precusor)

def screen_precusor_candidates(outputdir2, files, featureFile, precusorStrFile, rnafoldCmdTmp):
    prunedLength = 15
    closestDistance = 9
    fathestDistance = 44
    # The directory of featureFile, precusorFastaFile and precusorStrFile must be absolute directory.
    # miRNA will be tested by clusterSeq, majority seq and stable region of clusterSeq.
    dirTmp = outputdir2
    #os.path.dirname(featureFile)
    # Load the structure file into a dictionary
    wholePrecusorNameContentDic = {}
    with open(precusorStrFile, 'r') as inf2:
        totalContent = inf2.readlines()
    totalContentSplit = chunks(totalContent, 3)
    for item in totalContentSplit:
        precusorName = item[0].strip()[1:]
        precusorSeq = item[1].strip()
        precusorStr = item[2].strip().split()[0]
        try:
            if len(item[2].strip().split()[1]) >= 3:
                mfeValue = float(item[2].strip().split()[1][1:-1].strip())
            else:
                mfeValue = float(item[2].strip().split()[2][:-1].strip())
        except:
            mfeValue = 0.0
            print('Errors of mfe values happens at: %s. Please check it in %s'%(clusterName, precusorStrFile))
        wholePrecusorNameContentDic.update({precusorName:[precusorName, precusorSeq, precusorStr, mfeValue]})
    
    for clusterSeqType in ['stableClusterSeq']:
        clusterNameList = []
        miRNAlist = []
        neighborStateList = []
        upstreamDistanceList = []
        downstreamDistanceList = []
        clusterNameCandidatePreDic = {}
        clusterNameMiRNASeqDicTmp = {}
        clusterNameMiRNASeqDic = {}
        with open(featureFile , 'r') as inf1:
            line1 = inf1.readline()
            content1 = line1.strip().split('\t')
            miRNALabel = content1.index(clusterSeqType)
            clusterNameLabel = content1.index('clusterName')
            neighborStateLabel = content1.index('neighborState')
            upstreamDistanceLabel = content1.index('upstreamDistance')
            downstreamDistanceLabel = content1.index('downstreamDistance')
            line1 = inf1.readline()
            while line1 != '':
                content1 = line1.strip().split('\t')
                clusterName = content1[clusterNameLabel]
                miRNASeq = str(Seq(content1[miRNALabel]).transcribe())
                neighborState = content1[neighborStateLabel]
                if content1[upstreamDistanceLabel] == 'None':
                    upstreamDistance = 10000
                else:
                    upstreamDistance = int(content1[upstreamDistanceLabel])
                if content1[downstreamDistanceLabel] == 'None':
                    downstreamDistance = 10000
                else:
                    downstreamDistance = int(content1[downstreamDistanceLabel])
                clusterNameList.append(clusterName)
                miRNAlist.append(miRNASeq)
                neighborStateList.append(neighborState)
                upstreamDistanceList.append(upstreamDistance)
                downstreamDistanceList.append(downstreamDistance)
                clusterNameCandidatePreDic.update({clusterName:[]})
                clusterNameMiRNASeqDicTmp.update({clusterName:[]})
                line1 = inf1.readline()

        for i in range(len(clusterNameList)):
            clusterName = clusterNameList[i]
            direction = clusterName[-1]
            upstreamDistance = upstreamDistanceList[i]
            downstreamDistance = downstreamDistanceList[i]
            neighborState = neighborStateList[i]
            # precursor (70 bp upstream and 20 bp downstream)
            upPrecusor = wholePrecusorNameContentDic[clusterName+':precusor_1']
            # precursor (20 bp upstream and 70 bp downstream)
            downPrecusor = wholePrecusorNameContentDic[clusterName+':precusor_2']
            if neighborState != 'Good':
                clusterNameCandidatePreDic[clusterName].append(upPrecusor)
                clusterNameCandidatePreDic[clusterName].append(downPrecusor)
                clusterNameMiRNASeqDicTmp[clusterName].append(miRNAlist[i])
            else:
                if (upstreamDistance > fathestDistance or upstreamDistance < closestDistance) and (downstreamDistance <= fathestDistance and downstreamDistance >= closestDistance):
                    # The first one is the downPrecursor of this seq
                    clusterNameCandidatePreDic[clusterName].append(downPrecusor)
                    # The second one is the upPrecursor of the next seq 
                    clusterNameCandidatePreDic[clusterName].append(wholePrecusorNameContentDic[clusterNameList[i+1]+':precusor_1'])
                    clusterNameMiRNASeqDicTmp[clusterName].append(miRNAlist[i])
                    clusterNameMiRNASeqDicTmp[clusterName].append(miRNAlist[i+1])
                elif (upstreamDistance <= fathestDistance and upstreamDistance >=  closestDistance) and (downstreamDistance <= fathestDistance and downstreamDistance >= closestDistance):
                    # The first one is the downPrecursor of the previous seq
                    clusterNameCandidatePreDic[clusterName].append(wholePrecusorNameContentDic[clusterNameList[i-1]+':precusor_2'])
                    # The first one is the upPrecursor of this seq
                    clusterNameCandidatePreDic[clusterName].append(upPrecusor)
                    # The third one is the downPrecursor of this seq
                    clusterNameCandidatePreDic[clusterName].append(downPrecusor)
                    # The forth one is the upPrecursor of the next seq
                    clusterNameCandidatePreDic[clusterName].append(wholePrecusorNameContentDic[clusterNameList[i+1]+':precusor_1'])
                    clusterNameMiRNASeqDicTmp[clusterName].append(miRNAlist[i])
                    clusterNameMiRNASeqDicTmp[clusterName].append(miRNAlist[i-1])
                    clusterNameMiRNASeqDicTmp[clusterName].append(miRNAlist[i+1])
                elif (upstreamDistance <= fathestDistance and upstreamDistance >=  closestDistance) and (downstreamDistance > fathestDistance or downstreamDistance < closestDistance):
                    # The first one is the downPrecursor of the previous seq
                    clusterNameCandidatePreDic[clusterName].append(wholePrecusorNameContentDic[clusterNameList[i-1]+':precusor_2'])
                    # The second one is the upPrecursor of this seq
                    clusterNameCandidatePreDic[clusterName].append(upPrecusor)
                    clusterNameMiRNASeqDicTmp[clusterName].append(miRNAlist[i])
                    clusterNameMiRNASeqDicTmp[clusterName].append(miRNAlist[i-1])
                else:
                    print("Exception happens at: %s when picking candidate precursors"%(clusterName))

        for i in range(len(clusterNameList)):
            clusterName = clusterNameList[i]
            direction = clusterName[-1]
            candidatePreList = clusterNameCandidatePreDic[clusterName]
            clusterSeqListTmp = clusterNameMiRNASeqDicTmp[clusterName]
            clusterSeq = clusterSeqListTmp[0]
            numberOfCandidate = len(candidatePreList)
            if numberOfCandidate == 2:
                clusterSeqList = [clusterSeqListTmp for i in range(numberOfCandidate)]
            else:
                clusterSeqList = []
                clusterSeqList.append([clusterSeqListTmp[0], clusterSeqListTmp[1]])
                clusterSeqList.append([clusterSeqListTmp[0], clusterSeqListTmp[1]])
                clusterSeqList.append([clusterSeqListTmp[0], clusterSeqListTmp[2]])
                clusterSeqList.append([clusterSeqListTmp[0], clusterSeqListTmp[2]])
            clusterNameMiRNASeqDic.update({clusterName:clusterSeqList})

        # selcteOptimalPrecusor
        clustermiRNADic = {}
        clusterPrecusorDic = {}
        for i in range(len(clusterNameList)):
            clusterName = clusterNameList[i]
            precusorList = clusterNameCandidatePreDic[clusterName]
            precusorFeatureList = []
            prunedPrecusorList = []
            for index, precusor in enumerate(precusorList):
                precusorName = precusor[0]
                precusorSeq = precusor[1]
                precusorStr = precusor[2]
                precusorOrignialMFE = precusor[3]
                clusterSeqListTmp2 = clusterNameMiRNASeqDic[clusterName][index]
                if len(clusterSeqListTmp2) == 1:
                    precsorInstance = ReadPrecusor(rnafoldCmdTmp, precusorName, precusorSeq, precusorStr, precusorOrignialMFE, dirTmp, prunedLength, clusterSeqListTmp2[0])
                else:
                    precsorInstance = ReadPrecusor(rnafoldCmdTmp, precusorName, precusorSeq, precusorStr, precusorOrignialMFE, dirTmp, prunedLength, clusterSeqListTmp2[0], clusterSeqListTmp2[1])
                prunedPrecusorSeq, prunedPrecusorStr, prunedPrecusorMFE = precsorInstance.getPrunedPrecusor()
                if len(clusterSeqListTmp2) == 1:
                    prunedPrecursorInstance = ReadPrecusor(rnafoldCmdTmp, precusorName, prunedPrecusorSeq, prunedPrecusorStr, prunedPrecusorMFE, dirTmp, prunedLength, clusterSeqListTmp2[0])
                else:
                    prunedPrecursorInstance = ReadPrecusor(rnafoldCmdTmp, precusorName, prunedPrecusorSeq, prunedPrecusorStr, prunedPrecusorMFE, dirTmp, prunedLength, clusterSeqListTmp2[0], clusterSeqListTmp2[1])
                
                tmpList1 = []
                if not (prunedPrecusorStr is None)  and '(' in prunedPrecusorStr and ')' in prunedPrecusorStr:
                    percent = prunedPrecursorInstance.percentageOfPairedInMiRNA()
                    hairpinCount, bindingCount, interiorLoopCount, armType, apicalLoopSize, armTypeOverlenDic, stemLen, flag, mFE, pairState =  prunedPrecursorInstance.featuresInPrecusor()
                    precusorFeatureList.append([armTypeOverlenDic, hairpinCount, bindingCount, percent, mFE, precusorOrignialMFE])
                    tmpList1.append(precusorName)
                    tmpList1.append(prunedPrecusorSeq)
                    tmpList1.append(prunedPrecusorStr)
                    tmpList1.append(mFE)
                else:
                    precusorFeatureList.append([None, None, None, None, None, None])
                    tmpList1.append(precusorName)
                    tmpList1.append(prunedPrecusorSeq)
                    tmpList1.append(prunedPrecusorStr)
                    tmpList1.append(None)
                prunedPrecusorList.append(tmpList1)
            outputIndex, candiatePrecusor = selcteOptimalPrecusor(precusorFeatureList, prunedPrecusorList, clusterSeqType)
            if not (outputIndex is None):
                clustermiRNADic.update({clusterName:clusterNameMiRNASeqDic[clusterName][outputIndex]})
            if not (candiatePrecusor is None):
                clusterPrecusorDic.update({clusterName : candiatePrecusor})
        newFileHere = Path(outputdir2)/(files + '_updated_' + clusterSeqType + '_' + str(prunedLength) + '.tsv')
        outf = open(str(newFileHere), 'w')
        with open(featureFile , 'r') as inf:
            line = inf.readline()
            content1 = line.strip().split('\t')
            clusterNameLabel = content1.index('clusterName')
            outf.write(line.strip()+'\t')
            outf.write('pruned_precusor_seq\tpruned_precusor_str\tarmType\tdistanceToloop\tpercentage_PairedInMiRNA\thairpin_count\tbinding_count\tinteriorLoopCount\tapicalLoop_size\tstem_length\tmFE\tcount_bindings_in_miRNA\tUGU_UGUG_motif\tpair_state\n')
            line = inf.readline()
            while line != '':
                clusterName = line.strip().split('\t')[clusterNameLabel]
                if clusterName in clusterPrecusorDic.keys():
                    premiRNASeqContentList = clusterPrecusorDic[clusterName]
                    premiRNASeqName = premiRNASeqContentList[0]
                    premiRNASeq = premiRNASeqContentList[1]
                    premiRNAStructure = premiRNASeqContentList[2]
                    originalMFE = premiRNASeqContentList[3]
                    miRlist = clustermiRNADic[clusterName]
                    if len(miRlist) == 1:
                        miRNASeq = miRlist[0]
                        precsorInstance = ReadPrecusor(rnafoldCmdTmp, premiRNASeqName, premiRNASeq, premiRNAStructure, originalMFE, dirTmp, prunedLength, miRNASeq)
                    else:
                        miRNASeq1 = miRlist[0]
                        miRNASeq2 = miRlist[1]
                        precsorInstance = ReadPrecusor(rnafoldCmdTmp, premiRNASeqName, premiRNASeq, premiRNAStructure, originalMFE, dirTmp, prunedLength, miRNASeq1, miRNASeq2)
                    percent = precsorInstance.percentageOfPairedInMiRNA()
                    hairpinCount, bindingCount, interiorLoopCount, armType, apicalLoopSize, armTypeOverlenDic, stemLen, flag, mFE, pairState =  precsorInstance.featuresInPrecusor()
                    countBindingsInMiRNA = precsorInstance.bindingsInMiRNA()
                    outf.write(line.strip()+'\t')
                    outf.write('%s\t%s\t%s\t%d\t%.2f\t%d\t%d\t%d\t%d\t%d\t%.2f\t%d\t%s\t%s\n'%(premiRNASeq, premiRNAStructure, armType, armTypeOverlenDic[armType][1], percent, hairpinCount, bindingCount, interiorLoopCount, apicalLoopSize, stemLen, mFE, countBindingsInMiRNA, flag, pairState))
                else:
                    outf.write(line.strip()+'\t')
                    outf.write('\t'.join(['None']*14))
                    outf.write('\n')
                line = inf.readline()
        outf.close()



import os
import sys
import pandas as pd
import matplotlib
matplotlib.use('agg')
from sklearn.preprocessing import OneHotEncoder
import numpy as np
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from sklearn.feature_selection import VarianceThreshold, chi2, f_classif
from sklearn.model_selection import GridSearchCV
from sklearn.svm import SVC
from sklearn.metrics import make_scorer, confusion_matrix, matthews_corrcoef, roc_auc_score, roc_curve, auc
from sklearn.pipeline import Pipeline
import matplotlib.pyplot as plt
from pathlib import Path
import time
from matplotlib.backends.backend_pdf import PdfPages
import joblib

def preprocess_featureFiles(outputdir2, files, infTmp, feature_namelist_file):
    # infTmp is unmapped_mirna_JH-29_Pros_SMC_vs_representative_seq_modified_selected_sorted_features_updated_stableClusterSeq_15.tsv
    readCountLimit = 10
    seqCountLimit = 3
    tmpName = str(Path(outputdir2)/(files + '_dataset_15.csv'))
    #tmpName = '_'.join(os.path.basename(infTmp).split('_')[:-9])+'_dataset_'+os.path.basename(infTmp).split('_')[-1].split('.')[0]
    #with open(os.path.join(os.path.dirname(infTmp), tmpName+'.csv'), 'w') as outf:
    with open(tmpName, 'w') as outf:
        with open(infTmp, 'r') as inf:
            line = inf.readline()
        content = line.strip().split('\t')
        retainedIndexList = [0, 1, 5] + list(range(10, len(content)))
        retainedColNameList = [content[i] for i in range(len(content)) if i in retainedIndexList]
        readCountLimitLabel = content.index('readCountSum')
        seqCountLimitLabel = content.index('seqCount')

        with open(infTmp, 'r') as inf:
            line = inf.readline()
            content = line.strip().split('\t')
            outf.write(','.join([content[i] for i in range(len(content)) if i in retainedIndexList]))
            outf.write('\n')
            line = inf.readline()
            while line != '':
                content = line.strip().split('\t')
                if int(content[readCountLimitLabel]) >= readCountLimit and int(content[seqCountLimitLabel]) >= seqCountLimit and ('N' not in content[seqCountLimitLabel:]):
                    outf.write(','.join([content[i] for i in range(len(content)) if i in retainedIndexList]))
                    outf.write('\n')
                line = inf.readline()
            line = inf.readline()

    tmpName2 = str(Path(outputdir2)/(files + '_dataset_15_refined_tmp.csv'))
    outf = open(tmpName2, 'w')
    with open(tmpName, 'r') as inf:
        line = inf.readline()
        outf.write(line)
        content = line.strip().split(',')
        pruned_precusor_strLabel = content.index('pruned_precusor_str')
        line = inf.readline()
        while line != '':
            content = line.strip().split(',')
            if content[pruned_precusor_strLabel] != 'None':
                outf.write(line)
            else:
                pass
            line = inf.readline()
    outf.close()

    nameList = []
    with open(feature_namelist_file, 'r') as inf:
        for line in inf:
            if line.strip() not in nameList:
                nameList.append(line.strip())

    tmpName3 = str(Path(outputdir2)/(files + '_dataset_15_refined_features.csv'))
    outf = open(tmpName3, 'w')
    with open(tmpName2, 'r') as inf:
        retainedIndexList = []
        line = inf.readline()
        content = line.strip().split(',')
        reatinedContent = []
        for i, name in enumerate(content):
            if name in nameList:
                retainedIndexList.append(i)
                reatinedContent.append(name)
        outf.write(','.join(reatinedContent))
        outf.write('\n')
        line = inf.readline()
        while line != '':
            content = line.strip().split(',')
            reatinedContent = []
            for j in range(len(content)):
                if j in retainedIndexList:
                    reatinedContent.append(content[j])
            outf.write(','.join(reatinedContent))
            outf.write('\n')
            line = inf.readline()
    outf.close()


def model_predict(outputdir2, files, modelFile):
    inputFile = str(Path(outputdir2)/(files+"_dataset_15_refined_features.csv"))
    cutoff_picked = 0.8
    sc, clf, selectFeatureNameList = joblib.load(modelFile)
    data = pd.read_csv(inputFile)
    data_raw = data.iloc[:,:].values
    featureList_raw = data.columns.values.tolist()[:3]
    featureList_raw.insert(0, 'probability')
    featureList_raw.insert(0, 'predictedValue')
    # Encode categorical features using one-hot-coding after deleting the 2nd and 3rd column in data
    data.drop(data.columns[[0,1,2]], axis=1, inplace=True)
    data = pd.get_dummies(data)
    # Check the dummy varibles, such as 'pair_state_Yes', 'pair_state_No', etc.
    # If some dummy varibles are missing, they will be added with value of 0 across all of the samples.
    totalfeatureListTmp = data.columns.values.tolist()
    missedFeatureList = []
    for item in selectFeatureNameList:
        if item[2] not in totalfeatureListTmp:
            missedFeatureList.append(item[2])
    for item in missedFeatureList:
        data[item] = np.array([0]*(data.shape[0]))
    # Subselect data with the selected features and reorder the feature sequence according to the feature sequence in selectFeatureNameList
    totalfeatureList = data.columns.values.tolist()
    subIndexList = [totalfeatureList.index(item[2]) for item in selectFeatureNameList]
    
    data_x = data.iloc[:,subIndexList].values
    x_test_std_new = sc.transform(data_x)
    #featureLabel_new = subIndexList
    x_test_std_new2 = x_test_std_new
    predictedValue = clf.predict(x_test_std_new2)
    probability = [max(value) for value in clf.predict_proba(x_test_std_new2)]
    # Output the predicted result of the test data set into a file.
    array_predicted = np.column_stack((predictedValue, probability, data_raw[:,:3]))
    data_predicted_test = pd.DataFrame(array_predicted, columns=featureList_raw)
    data_predicted_test.to_csv(str(Path(outputdir2)/('predicted_'+files+'_detailed.csv')), index=False)
    outf = open(str(Path(outputdir2)/(files+'_novel_miRNAs_miRge2.0.csv')),'w')
    contentList_tmp = []
    with open(str(Path(outputdir2)/('predicted_'+files+'_detailed.csv')), 'r') as inf:
        line = inf.readline()
        outf.write(line)
        line = inf.readline()
        while line != '':
            content = line.strip().split(',')
            if content[0] == '1' and float(content[1]) >= cutoff_picked:
                contentList_tmp.append([float(content[1]), line])
            line = inf.readline()
    contentList_tmp.sort(reverse=True)
    for item in contentList_tmp:
        outf.write(item[1])
    outf.close()



#!/usr/bin/python

#import timeit, sys

import numpy as np
import math as m

def magnitude(vec):
    '''
    Return the magnitude of a vector (|V|).

    @param vec: The vector in question.
    @return: The magnitude of the vector.
    '''
    #return ftuc.magnitude(vec)
    return m.sqrt(np.dot(vec, vec))

def vec_distance(vec1, vec2):
    #return ftuc.vec_distance(vec1, vec2)
    return m.sqrt(np.dot(vec2 - vec1, vec2 - vec1))
